#!/usr/bin/env python3

# Copyright © 2020-2024, Meheret Tesfaye Batu <meherett.batu@gmail.com>
#             2024, Eyoel Tadesse <eyoel_tadesse@proton.me>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from hdwallet.symbols import *


def test_symbols():

	assert ACC == "ACC"
	assert AKT == "AKT"
	assert ALGO == "ALGO"
	assert ANON == "ANON"
	assert APT == "APT"
	assert ARB == "ARB"
	assert AGM == "AGM"
	assert XAX == "XAX"
	assert AYA == "AYA"
	assert AC == "AC"
	assert AUR == "AUR"
	assert AVAX == "AVAX"
	assert AVN == "AVN"
	assert AXE == "AXE"
	assert AXL == "AXL"
	assert BAND == "BAND"
	assert BTA == "BTA"
	assert BEET == "BEET"
	assert BELA == "BELA"
	assert BNB == "BNB"
	assert BTDX == "BTDX"
	assert BTC == "BTC"
	assert BCA == "BCA"
	assert BCH == "BCH"
	assert SLP == "SLP"
	assert BTG == "BTG"
	assert BITG == "BITG"
	assert XBC == "XBC"
	assert BTCP == "BTCP"
	assert BSV == "BSV"
	assert BTCZ == "BTCZ"
	assert BTX == "BTX"
	assert BSD == "BSD"
	assert BLK == "BLK"
	assert BND == "BND"
	assert BST == "BST"
	assert BOLI == "BOLI"
	assert BRIT == "BRIT"
	assert CDN == "CDN"
	assert CCN == "CCN"
	assert ADA == "ADA"
	assert CELO == "CELO"
	assert HUA == "HUA"
	assert CLAM == "CLAM"
	assert CLUB == "CLUB"
	assert CMP == "CMP"
	assert ATOM == "ATOM"
	assert CPU == "CPU"
	assert CRP == "CRP"
	assert CRAVE == "CRAVE"
	assert DASH == "DASH"
	assert ONION == "ONION"
	assert DFC == "DFC"
	assert DNR == "DNR"
	assert DMD == "DMD"
	assert DGB == "DGB"
	assert DGC == "DGC"
	assert DIVI == "DIVI"
	assert DOGE == "DOGE"
	assert DYDX == "DYDX"
	assert XEC == "XEC"
	assert ECN == "ECN"
	assert EDRC == "EDRC"
	assert EFL == "EFL"
	assert EMC2 == "EMC2"
	assert ELA == "ELA"
	assert NRG == "NRG"
	assert EOS == "EOS"
	assert ERG == "ERG"
	assert ETH == "ETH"
	assert ERC == "ERC"
	assert EVR == "EVR"
	assert EXCL == "EXCL"
	assert FTM == "FTM"
	assert FTC == "FTC"
	assert FET == "FET"
	assert FIL == "FIL"
	assert FIRO == "FIRO"
	assert FRST == "FRST"
	assert FIX == "FIX"
	assert FLASH == "FLASH"
	assert FLUX == "FLUX"
	assert FOXD == "FOXD"
	assert FJC == "FJC"
	assert GAME == "GAME"
	assert GCR == "GCR"
	assert GBX == "GBX"
	assert GRC == "GRC"
	assert GRS == "GRS"
	assert NLG == "NLG"
	assert ONE == "ONE"
	assert HNC == "HNC"
	assert THC == "THC"
	assert ZEN == "ZEN"
	assert HT == "HT"
	assert HUSH == "HUSH"
	assert ICX == "ICX"
	assert INJ == "INJ"
	assert INSN == "INSN"
	assert IOP == "IOP"
	assert IRIS == "IRIS"
	assert IXC == "IXC"
	assert JBS == "JBS"
	assert KAVA == "KAVA"
	assert KOBO == "KOBO"
	assert KMD == "KMD"
	assert LDCN == "LDCN"
	assert LBC == "LBC"
	assert LINX == "LINX"
	assert LTC == "LTC"
	assert LCC == "LCC"
	assert LTZ == "LTZ"
	assert LKR == "LKR"
	assert LYNX == "LYNX"
	assert MZC == "MZC"
	assert MEC == "MEC"
	assert METIS == "METIS"
	assert MNX == "MNX"
	assert MONA == "MONA"
	assert XMR == "XMR"
	assert MONK == "MONK"
	assert EGLD == "EGLD"
	assert XMY == "XMY"
	assert NMC == "NMC"
	assert XNO == "XNO"
	assert NAV == "NAV"
	assert NEAR == "NEAR"
	assert NEBL == "NEBL"
	assert NEO == "NEO"
	assert NEOS == "NEOS"
	assert NRO == "NRO"
	assert NTRN == "NTRN"
	assert NYC == "NYC"
	assert NCG == "NCG"
	assert NIX == "NIX"
	assert NVC == "NVC"
	assert NBT == "NBT"
	assert NSR == "NSR"
	assert OK == "OK"
	assert OKT == "OKT"
	assert OMNI == "OMNI"
	assert ONX == "ONX"
	assert ONT == "ONT"
	assert OP == "OP"
	assert OSMO == "OSMO"
	assert PART == "PART"
	assert PPC == "PPC"
	assert PSB == "PSB"
	assert PHR == "PHR"
	assert PI == "PI"
	assert PINK == "PINK"
	assert PIVX == "PIVX"
	assert MATIC == "MATIC"
	assert POSW == "POSW"
	assert POT == "POT"
	assert PRJ == "PRJ"
	assert PUT == "PUT"
	assert QTUM == "QTUM"
	assert RPD == "RPD"
	assert RVN == "RVN"
	assert RDD == "RDD"
	assert XRP == "XRP"
	assert RITO == "RITO"
	assert RBTC == "RBTC"
	assert RBY == "RBY"
	assert SAFE == "SAFE"
	assert SLS == "SLS"
	assert SCRIBE == "SCRIBE"
	assert SCRT == "SCRT"
	assert SDC == "SDC"
	assert CTK == "CTK"
	assert SLM == "SLM"
	assert SMLY == "SMLY"
	assert SOL == "SOL"
	assert SLR == "SLR"
	assert FIS == "FIS"
	assert STASH == "STASH"
	assert XLM == "XLM"
	assert STRAT == "STRAT"
	assert SUGAR == "SUGAR"
	assert SUI == "SUI"
	assert SYS == "SYS"
	assert LUNA == "LUNA"
	assert XTZ == "XTZ"
	assert THETA == "THETA"
	assert THT == "THT"
	assert TOA == "TOA"
	assert TRX == "TRX"
	assert TWINS == "TWINS"
	assert USC == "USC"
	assert UNO == "UNO"
	assert VC == "VC"
	assert VET == "VET"
	assert XVG == "XVG"
	assert VTC == "VTC"
	assert VIA == "VIA"
	assert VIVO == "VIVO"
	assert VOX == "VOX"
	assert VASH == "VASH"
	assert WGR == "WGR"
	assert XWC == "XWC"
	assert WC == "WC"
	assert XDC == "XDC"
	assert XUEZ == "XUEZ"
	assert YEC == "YEC"
	assert ZEC == "ZEC"
	assert ZCL == "ZCL"
	assert ZET == "ZET"
	assert ZIL == "ZIL"
	assert ZBC == "ZBC"



#!/usr/bin/env python3

# Copyright © 2020-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from ...iecc import IEllipticCurveCryptography
from .point import SLIP10Ed25519Point
from .public_key import SLIP10Ed25519PublicKey
from .private_key import SLIP10Ed25519PrivateKey


class SLIP10Ed25519ECC(IEllipticCurveCryptography):

    NAME = "SLIP10-Ed25519"
    ORDER = 2 ** 252 + 27742317777372353535851937790883648493
    GENERATOR = SLIP10Ed25519Point.from_coordinates(
        15112221349535400772501151409588531511454012693041857206046113283949847762202,
        46316835694926478169428394003475163141307993866256225615783033603165251855960
    )
    POINT = SLIP10Ed25519Point
    PUBLIC_KEY = SLIP10Ed25519PublicKey
    PRIVATE_KEY = SLIP10Ed25519PrivateKey



#!/usr/bin/env python3

# Copyright © 2020-2024, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

import json
import click
import sys

from ...entropies import (
    IEntropy,
    AlgorandEntropy, ALGORAND_ENTROPY_STRENGTHS,
    BIP39Entropy, BIP39_ENTROPY_STRENGTHS,
    ElectrumV1Entropy, ELECTRUM_V1_ENTROPY_STRENGTHS,
    ElectrumV2Entropy, ELECTRUM_V2_ENTROPY_STRENGTHS,
    MoneroEntropy, MONERO_ENTROPY_STRENGTHS,
    ENTROPIES
)


def generate_entropy(**kwargs) -> None:
    try:
        if not ENTROPIES.is_entropy(name=kwargs.get("client")):
            click.echo(click.style(
                f"Wrong entropy client, (expected={ENTROPIES.names()}, got='{kwargs.get('client')}')"
            ), err=True)
            sys.exit()

        if kwargs.get("strength") is None:  # Set default strength
            if kwargs.get("client") == AlgorandEntropy.name():
                strength: int = ALGORAND_ENTROPY_STRENGTHS.TWO_HUNDRED_FIFTY_SIX
            elif kwargs.get("client") == BIP39Entropy.name():
                strength: int = BIP39_ENTROPY_STRENGTHS.ONE_HUNDRED_TWENTY_EIGHT
            elif kwargs.get("client") == ElectrumV1Entropy.name():
                strength: int = ELECTRUM_V1_ENTROPY_STRENGTHS.ONE_HUNDRED_TWENTY_EIGHT
            elif kwargs.get("client") == ElectrumV2Entropy.name():
                strength: int = ELECTRUM_V2_ENTROPY_STRENGTHS.ONE_HUNDRED_THIRTY_TWO
            elif kwargs.get("client") == MoneroEntropy.name():
                strength: int = MONERO_ENTROPY_STRENGTHS.ONE_HUNDRED_TWENTY_EIGHT
        else:
            strength: int = kwargs.get("strength")

        if not ENTROPIES.entropy(name=kwargs.get("client")).is_valid_strength(strength=strength):
            click.echo(click.style(
                f"Wrong {kwargs.get('client')} entropy strength, "
                f"(expected={ENTROPIES.entropy(name=kwargs.get('client')).strengths}, got='{strength}')"
            ), err=True)
            sys.exit()

        entropy: IEntropy = ENTROPIES.entropy(name=kwargs.get("client")).__call__(
            entropy=ENTROPIES.entropy(name=kwargs.get("client")).generate(
                strength=strength
            )
        )
        click.echo(json.dumps(
            {
                "client": entropy.name(),
                "entropy": entropy.entropy(),
                "strength": entropy.strength()
            },
            indent=kwargs.get("indent", 4),
            ensure_ascii=kwargs.get("ensure_ascii", False)
        ))

    except Exception as exception:
        click.echo(click.style(f"Error: {str(exception)}"), err=True)
        sys.exit()



#!/usr/bin/env python3

# Copyright © 2020-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import (
    Any, Union
)

from ..libs.base58 import (
    ensure_string, check_encode, check_decode
)
from ..eccs import (
    IPublicKey, SLIP10Nist256p1PublicKey, validate_and_get_public_key
)
from ..cryptocurrencies import Neo
from ..crypto import hash160
from ..utils import (
    integer_to_bytes, bytes_to_string
)
from .iaddress import IAddress


class NeoAddress(IAddress):

    address_prefix: bytes = integer_to_bytes(Neo.PARAMS.ADDRESS_PREFIX)
    address_suffix: bytes = integer_to_bytes(Neo.PARAMS.ADDRESS_SUFFIX)
    address_version: bytes = integer_to_bytes(Neo.PARAMS.ADDRESS_VERSION)

    alphabet: bytes = Neo.PARAMS.ALPHABET

    @staticmethod
    def name() -> str:
        """
        Returns the name of the cryptocurrency, which is "Neo".

        :return: The name of the cryptocurrency.
        :rtype: str
        """

        return "Neo"

    @classmethod
    def encode(cls, public_key: Union[bytes, str, IPublicKey], **kwargs: Any) -> str:
        """
        Encode a public key into a cryptocurrency address using SLIP-10 NIST P-256 public key.

        :param public_key: The public key to encode. Can be bytes, str, or IPublicKey.
        :type public_key: Union[bytes, str, IPublicKey]
        :param kwargs: Additional keyword arguments.
            - address_version: Version of the address (optional).
            - alphabet: Custom alphabet for encoding (optional).
        :type kwargs: Any

        :return: The encoded cryptocurrency address.
        :rtype: str
        """

        public_key: IPublicKey = validate_and_get_public_key(
            public_key=public_key, public_key_cls=SLIP10Nist256p1PublicKey
        )
        payload: bytes = (
            cls.address_prefix + public_key.raw_compressed() + cls.address_suffix
        )
        payload_hash: bytes = hash160(payload)

        return ensure_string(check_encode(
            (kwargs.get("address_version", cls.address_version) + payload_hash), alphabet=kwargs.get(
                "alphabet", cls.alphabet
            )
        ))

    @classmethod
    def decode(cls, address: str, **kwargs: Any) -> str:
        """
        Decode a cryptocurrency address into its public key using SLIP-10 NIST P-256 public key.

        :param address: The address to decode.
        :type address: str
        :param kwargs: Additional keyword arguments.
            - alphabet: Custom alphabet for decoding (optional).
            - address_version: Version of the address (optional).
        :type kwargs: Any

        :return: The decoded public key.
        :rtype: str
        """

        address_decode: bytes = check_decode(
            address, alphabet=kwargs.get(
                "alphabet", cls.alphabet
            )
        )
        address_version: bytes = kwargs.get("address_version", cls.address_version)

        expected_length: int = 20 + len(address_version)
        if len(address_decode) != expected_length:
            raise ValueError(f"Invalid length (expected: {expected_length}, got: {len(address_decode)})")

        address_version_got: bytes = integer_to_bytes(address_decode[0])
        if address_version != address_version_got:
            raise ValueError(f"Invalid address version (expected: {address_version}, got: {address_version_got})")

        return bytes_to_string(address_decode[1:])



#!/usr/bin/env python3

# Copyright © 2020-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import (
    Any, Union
)

from ..libs.bech32 import (
    bech32_encode, bech32_decode
)
from ..eccs import (
    IPublicKey, SLIP10Secp256k1PublicKey, validate_and_get_public_key
)
from ..cryptocurrencies import Zilliqa
from ..crypto import sha256
from ..utils import bytes_to_string
from .iaddress import IAddress


class ZilliqaAddress(IAddress):

    hrp: str = Zilliqa.NETWORKS.MAINNET.HRP

    @staticmethod
    def name() -> str:
        """
        Return the name associated with the Zilliqa blockchain.

        :return: The name "Zilliqa".
        :rtype: str
        """

        return "Zilliqa"

    @classmethod
    def encode(cls, public_key: Union[bytes, str, IPublicKey], **kwargs: Any) -> str:
        """
        Encode a public key into a Zilliqa address using Bech32 encoding.

        :param public_key: The public key to encode, which can be bytes, str, or an IPublicKey object.
        :type public_key: Union[bytes, str, IPublicKey]
        :param kwargs: Additional keyword arguments:
            - hrp (str, optional): Human-readable part for the Bech32 encoding. Defaults to cls.hrp.

        :return: Encoded Zilliqa address.
        :rtype: str
        """

        public_key: IPublicKey = validate_and_get_public_key(
            public_key=public_key, public_key_cls=SLIP10Secp256k1PublicKey
        )
        public_key_hash: bytes = sha256(public_key.raw_compressed())

        return bech32_encode(
            kwargs.get("hrp", cls.hrp), public_key_hash[-20:]
        )

    @classmethod
    def decode(cls, address: str, **kwargs: Any) -> str:
        """
        Decode a Zilliqa address from Bech32 encoding.

        :param address: The Zilliqa address to decode.
        :type address: str
        :param kwargs: Additional keyword arguments:
            - hrp (str, optional): Human-readable part for the Bech32 encoding. Defaults to cls.hrp.

        :return: Decoded Zilliqa address.
        :rtype: str
        """

        hrp, address_decode = bech32_decode(
            kwargs.get("hrp", cls.hrp), address
        )

        if len(address_decode) != 20:
            raise ValueError(f"Invalid length (expected: {20}, got: {len(address_decode)})")

        return bytes_to_string(address_decode)



#!/usr/bin/env python3

# Copyright © 2020-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import Any
from ecdsa import VerifyingKey
from ecdsa.ecdsa import curve_256
from ecdsa import (
    curves, ellipticcurve, keys
)

from ....consts import SLIP10_SECP256K1_CONST
from ...iecc import (
    IPoint, IPublicKey
)
from .point import SLIP10Nist256p1Point


class SLIP10Nist256p1PublicKey(IPublicKey):
    
    verify_key: VerifyingKey

    def __init__(self, verify_key: VerifyingKey) -> None:
        """
        Initialize the public key with the given verifying key.

        :param verify_key: The verifying key used for initializing the public key.
        :type verify_key: VerifyingKey
        """

        self.verify_key = verify_key

    @staticmethod
    def name() -> str:
        """
        Get the name of the ecc class.

        :return: The name of the ecc class.
        :rtype: str
        """

        return "SLIP10-Nist256p1"

    @classmethod
    def from_bytes(cls, public_key: bytes) -> IPublicKey:
        """
        Create a public key instance from the given byte representation.

        :param public_key: The byte representation of the public key.
        :type public_key: bytes

        :return: An instance of IPublicKey.
        :rtype: IPublicKey
        """

        try:
            return cls(
                VerifyingKey.from_string(
                    public_key, curve=curves.NIST256p
                )
            )
        except keys.MalformedPointError as ex:
            raise ValueError("Invalid public key bytes") from ex

    @classmethod
    def from_point(cls, point: IPoint) -> IPublicKey:
        """
        Create a public key instance from a given point.

        :param point: The elliptic curve point representing the public key.
        :type point: IPoint

        :return: An instance of IPublicKey.
        :rtype: IPublicKey
        """

        try:
            return cls(
                VerifyingKey.from_public_point(
                    ellipticcurve.Point(
                        curve_256, point.x(), point.y()
                    ),
                    curve=curves.NIST256p
                )
            )
        except keys.MalformedPointError as ex:
            raise ValueError("Invalid public key point") from ex

    @staticmethod
    def compressed_length() -> int:
        """
        Get the length of the compressed public key in bytes.

        :return: The length of the compressed public key.
        :rtype: int
        """

        return SLIP10_SECP256K1_CONST.PUBLIC_KEY_COMPRESSED_BYTE_LENGTH

    @staticmethod
    def uncompressed_length() -> int:
        """
        Get the length of the uncompressed public key in bytes.

        :return: The length of the uncompressed public key.
        :rtype: int
        """

        return SLIP10_SECP256K1_CONST.PUBLIC_KEY_UNCOMPRESSED_BYTE_LENGTH

    def underlying_object(self) -> Any:
        """
        Get the underlying object of the public key.

        :return: The underlying object representing the public key.
        :rtype: Any
        """

        return self.verify_key

    def raw_compressed(self) -> bytes:
        """
        Get the compressed raw bytes representation of the public key.

        :return: The compressed raw bytes of the public key.
        :rtype: bytes
        """

        return self.verify_key.to_string("compressed")

    def raw_uncompressed(self) -> bytes:
        """
        Get the uncompressed raw bytes representation of the public key.

        :return: The uncompressed raw bytes of the public key.
        :rtype: bytes
        """

        return self.verify_key.to_string("uncompressed")

    def point(self) -> IPoint:
        """
        Get the elliptic curve point corresponding to the public key.

        :return: The elliptic curve point.
        :rtype: IPoint
        """

        return SLIP10Nist256p1Point(self.verify_key.pubkey.point)



#!/usr/bin/env python3

# Copyright © 2020-2024, Meheret Tesfaye Batu <meherett.batu@gmail.com>
#             2024, Eyoel Tadesse <eyoel_tadesse@proton.me>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

import pytest
from hdwallet.crypto import (
    hmac_sha256, hmac_sha512, blake2b, blake2b_32, blake2b_40, blake2b_160, blake2b_224, blake2b_256, blake2b_512,
    chacha20_poly1305_encrypt, chacha20_poly1305_decrypt, sha256, double_sha256, hash160, crc32, xmodem_crc, 
    pbkdf2_hmac_sha512, kekkak256, ripemd160, sha512, sha512_256, sha3_256
)

# def test_hmac_sha256():
#     assert hmac_sha256("key", "data") == b'todo_mock'
#     assert hmac_sha256(b"key", b"data") == b'todo_mock'

# def test_hmac_sha512():
#     assert hmac_sha512("key", "data") == b'todo_mock'
#     assert hmac_sha512(b"key", b"data") == b'todo_mock'

# def test_blake2b():
#     assert blake2b("data", 64) == b'todo_mock'

# def test_blake2b_32():
#     assert blake2b_32("data") == b'todo_mock'

# def test_blake2b_40():
#     assert blake2b_40("data") == b'todo_mock'

# def test_blake2b_160():
#     assert blake2b_160("data") == b'todo_mock'

# def test_blake2b_224():
#     assert blake2b_224("data") == b'todo_mock'

# def test_blake2b_256():
#     assert blake2b_256("data") == b'todo_mock'

# def test_blake2b_512():
#     assert blake2b_512("data") == b'todo_mock'

# def test_chacha20_poly1305_encrypt_decrypt():
#     key = b'0' * 32
#     nonce = b'0' * 12
#     assoc_data = b'header'
#     plain_text = b'plaintext'
#     cipher_text, tag = chacha20_poly1305_encrypt(key, nonce, assoc_data, plain_text)
#     decrypted_text = chacha20_poly1305_decrypt(key, nonce, assoc_data, cipher_text, tag)
#     assert decrypted_text == plain_text

# def test_sha256():
#     assert sha256("data") == b'todo_mock'

# def test_double_sha256():
#     assert double_sha256("data") == b'todo_mock'

# def test_hash160():
#     assert hash160("data") == b'todo_mock'

# def test_crc32():
#     assert crc32("data") == b'todo_mock'

# def test_xmodem_crc():
#     assert xmodem_crc("data") == b'todo_mock'

# def test_pbkdf2_hmac_sha512():
#     assert pbkdf2_hmac_sha512("password", "salt", 1000) == b'todo_mock'

# def test_kekkak256():
#     assert kekkak256("data") == b'todo_mock'

# def test_ripemd160():
#     assert ripemd160("data") == b'todo_mock'

# def test_sha512():
#     assert sha512("data") == b'todo_mock'

# def test_sha512_256():
#     assert sha512_256("data") == b'todo_mock'

# def test_sha3_256():
#     assert sha3_256("data") == b'todo_mock'



#!/usr/bin/env python3

# Copyright © 2020-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from abc import (
    ABC, abstractmethod
)
from typing import Union

import re

from ..mnemonics import IMnemonic


class ISeed(ABC):

    _name: str
    _seed: str

    length: int

    def __init__(self, seed: str, **kwargs) -> None:
        """
        Initialize an object with a seed value.

        :param seed: The seed value used for initialization.
        :type seed: str

        :return: No return
        :rtype: NoneType
        """

        self._seed = seed

    @classmethod
    def name(cls) -> str:
        pass

    @classmethod
    def is_valid(cls, seed: str) -> bool:
        """
        Checks if the given seed is valid.

        :param seed: Hex string representing seed
        :type seed: str

        :return: True if is valid, False otherwise.
        :rtype: bool
        """

        return isinstance(seed, str) and bool(re.fullmatch(
            r'^[0-9a-fA-F]+$', seed
        )) and len(seed) == cls.length

    def seed(self) -> str:
        """
        Retrieves the seed associated with the current instance.

        :return: The seed as a string.
        :rtype: str
        """

        return self._seed

    @classmethod
    @abstractmethod
    def from_mnemonic(cls, mnemonic: Union[str, IMnemonic], **kwargs) -> str:
        pass



#!/usr/bin/env python3

# Copyright © 2020-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import Union

from ...crypto import sha256
from ...exceptions import MnemonicError
from ...mnemonics import (
    IMnemonic, ElectrumV1Mnemonic
)
from ...utils import bytes_to_string, encode
from ..iseed import ISeed


class ElectrumV1Seed(ISeed):
    """
    This class generates a root extended private key from a given seed using the
    Electrum-V1 standard. The Electrum-V1 standard defines a method for generating mnemonic
    phrases and converting them into a binary seed used for hierarchical
    deterministic wallets.

    .. note::
        This class inherits from the ``ISeed`` class, thereby ensuring that all functions are accessible.
    """

    hash_iteration_number: int = 10 ** 5

    length = 64

    @classmethod
    def name(cls) -> str:
        """
        Get the name of the seeds class.

        :return: The name of the seeds class.
        :rtype: str
        """

        return "Electrum-V1"

    @classmethod
    def from_mnemonic(cls, mnemonic: Union[str, IMnemonic], **kwargs) -> str:
        """
        Converts an Electrum V1 mnemonic phrase to its corresponding hashed entropy.

        :param mnemonic: The mnemonic phrase to be decoded. Can be a string or an instance of `IMnemonic`.
        :type mnemonic: Union[str, IMnemonic]

        :return: The hashed entropy as a string.
        :rtype: str
        """

        mnemonic = (
            mnemonic.mnemonic() if isinstance(mnemonic, IMnemonic) else mnemonic
        )
        if not ElectrumV1Mnemonic.is_valid(mnemonic=mnemonic):
            raise MnemonicError(f"Invalid {cls.name()} mnemonic words")

        entropy: str = ElectrumV1Mnemonic.decode(mnemonic)
        entropy_hash: bytes = encode(entropy)
        for _ in range(cls.hash_iteration_number):
            entropy_hash = sha256(entropy_hash + encode(entropy))

        return bytes_to_string(entropy_hash)



#!/usr/bin/env python3

# Copyright © 2020-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from click_aliases import ClickAliasedGroup

import click

from ..cryptocurrencies import Cardano
from ..mnemonics import ELECTRUM_V2_MNEMONIC_TYPES
from ..consts import (
    PUBLIC_KEY_TYPES, MODES
)

from .. import __version__

from .generate.entropy import generate_entropy
from .generate.mnemonic import generate_mnemonic
from .generate.seed import generate_seed
from .dump import dump
from .dumps import dumps
from .list.cryptocurrencies import list_cryptocurrencies
from .list.languages import list_languages
from .list.strengths import list_strengths


def current_version(
    context: click.core.Context, option: click.core.Option, value: bool
) -> None:
    if not value or context.resilient_parsing:
        return
    click.echo(__version__)
    context.exit()


@click.group(
    cls=ClickAliasedGroup,
    options_metavar="[OPTIONS]",
    context_settings={
        "help_option_names": ["-h", "--help"]
    }
)
@click.option(
    "-v", "--version",
    is_flag=True,
    callback=current_version,
    expose_value=False,
    help="Show HDWallet version and exit"
)
def cli_main():
    pass


@cli_main.group(
    "generate",
    aliases=["g"],
    cls=ClickAliasedGroup,
    options_metavar="[OPTIONS]",
    short_help="Select Generate for HDWallet",
    invoke_without_command=True
)
@click.pass_context
def generate(context: click.core.Context) -> None:
    if context.invoked_subcommand is None:
        pass


@generate.command(
    "entropy",
    aliases=["e"],
    options_metavar="[OPTIONS]",
    short_help="Select Entropy for generation entropy"
)
@click.option(
    "-c", "--client", type=str, default="BIP39", help="Set Entropy client", show_default=True
)
@click.option(
    "-s", "--strength", type=int, default=None, help="Set Strength for entropy", show_default=True
)
def cli_entropy(**kwargs) -> None:
    return generate_entropy(**kwargs)


@generate.command(
    "mnemonic",
    aliases=["m"],
    options_metavar="[OPTIONS]",
    short_help="Select Mnemonic for generation mnemonic"
)
@click.option(
    "-c", "--client", type=str, default="BIP39", help="Set Mnemonic client", show_default=True
)
@click.option(
    "-l", "--language", type=str, default=None, help="Set Mnemonic language", show_default=True
)
@click.option(
    "-e", "--entropy", type=str, default=None, help="Set Mnemonic entropy", show_default=True
)
@click.option(
    "-w", "--words", type=int, default=None, help="Set Mnemonic words", show_default=True
)
@click.option(
    "-mt", "--mnemonic-type", type=str, default="standard", help="Set Mnemonic type for Electrum-V2", show_default=True
)
@click.option(
    "-max", "--max-attempts", type=int, default=(10 ** 60), help="Set Max attempts for Electrum-V2", show_default=True
)
@click.option(
    "-cs", "--checksum", type=bool, default=False, help="Set Checksum for Monero", show_default=True
)
def cli_mnemonic(**kwargs) -> None:
    return generate_mnemonic(**kwargs)


@generate.command(
    "seed",
    aliases=["s"],
    options_metavar="[OPTIONS]",
    short_help="Select Seed for generation seed"
)
@click.option(
    "-c", "--client", type=str, default="BIP39", help="Set Seed client", show_default=True
)
@click.option(
    "-m", "--mnemonic", type=str, default=None, help="Set Seed mnemonic", show_default=True
)
@click.option(
    "-p", "--passphrase", type=str, default=None, help="Set Seed passphrase", show_default=True
)
@click.option(
    "-ct", "--cardano-type", type=str, default=Cardano.TYPES.BYRON_ICARUS, help="Set Cardano Seed type", show_default=True
)
@click.option(
    "-mt", "--mnemonic-type", type=str, default="standard", help="Set Mnemonic type for Electrum-V2", show_default=True
)
def cli_seed(**kwargs) -> None:
    return generate_seed(**kwargs)


@cli_main.command(
    "dump", aliases=["d"], options_metavar="[OPTIONS]", short_help="Select Dump hdwallet keys"
)
@click.option(
    "-sym", "--symbol", type=str, default="BTC", help="Set Cryptocurrency ticker symbol", show_default=True
)
@click.option(
    "-h", "--hd", type=str, default="BIP44", help="Select HD", show_default=True
)
@click.option(
    "-n", "--network", type=str, default="mainnet", help="Select Network type", show_default=True
)
@click.option(
    "-ec", "--entropy-client", type=str, default="BIP39", help="Select Entropy client", show_default=True
)
@click.option(
    "-e", "--entropy", type=str, default=None, help="Set Master key from Entropy hex string", show_default=True
)
@click.option(
    "-mc", "--mnemonic-client", type=str, default="BIP39", help="Select Mnemonic client", show_default=True
)
@click.option(
    "-m", "--mnemonic", type=str, default=None, help="Set Master key from Mnemonic words", show_default=True
)
@click.option(
    "-l", "--language", type=str, default="english", help="Select Language for mnemonic", show_default=True
)
@click.option(
    "-sc", "--seed-client", type=str, default="BIP39", help="Select Seed client", show_default=True
)
@click.option(
    "-s", "--seed", type=str, default=None, help="Set Master key from Seed hex string", show_default=True
)
@click.option(
    "-pp", "--passphrase", type=str, default=None, help="Set Passphrase for mnemonic & seed", show_default=True
)
@click.option(
    "-xprv", "--xprivate-key", type=str, default=None, help="Set Master key from XPrivate key", show_default=True
)
@click.option(
    "-xpub", "--xpublic-key", type=str, default=None, help="Set Master key from XPublic key", show_default=True
)
@click.option(
    "-enc", "--encoded", type=bool, default=True, help="Set Encoded for XPrivate & XPublic keys", show_default=True
)
@click.option(
    "-st", "--strict", type=bool, default=False, help="Set Strict for Root XPrivate & XPublic keys", show_default=True
)
@click.option(
    "-pkt", "--public-key-type", type=str, default=PUBLIC_KEY_TYPES.COMPRESSED, help="Select Public key type", show_default=True
)
@click.option(
    "-d", "--derivation", type=str, default="BIP44", help="Select Derivation name", show_default=True
)
@click.option(
    "-ac", "--account", type=str, default="0", help="Set Account index for derivation", show_default=True
)
@click.option(
    "-ch", "--change", type=str, default="0", help="Set Change index for derivation", show_default=True
)
@click.option(
    "-ecc", "--ecc", type=str, default="0", help="Set ECC index for HDW derivation", show_default=True
)
@click.option(
    "-ro", "--role", type=str, default="0", help="Set Role index for CIP1852 derivation", show_default=True
)
@click.option(
    "-ad", "--address", type=str, default="0", help="Set Address index for derivation", show_default=True
)
@click.option(
    "-mi", "--minor", type=str, default="1", help="Set Minor index for Monero derivation", show_default=True
)
@click.option(
    "-ma", "--major", type=str, default="0", help="Set Major index for Monero derivation", show_default=True
)
@click.option(
    "-p", "--path", type=str, default=None, help="Set Path for derivation", show_default=True
)
@click.option(
    "-i", "--indexes", type=list, default=[], help="Set Indexes for derivation", show_default=True
)
@click.option(
    "-prv", "--private-key", type=str, default=None, help="Set Private key", show_default=True
)
@click.option(
    "-w", "--wif", type=str, default=None, help="Set Wallet Import Format (WIF)", show_default=True
)
@click.option(
    "-b38", "--bip38", type=bool, default=False, help="Is BIP38 Encrypted Wallet Import Format", show_default=True
)
@click.option(
    "-pub", "--public-key", type=str, default=None, help="Set Public key", show_default=True
)
@click.option(
    "-sprv", "--spend-private-key", type=str, default=None, help="Set Spend Private key for Monero", show_default=True
)
@click.option(
    "-vprv", "--view-private-key", type=str, default=None, help="Set View Private key for Monero", show_default=True
)
@click.option(
    "-spub", "--spend-public-key", type=str, default=None, help="Set Spend Public key for Monero", show_default=True
)
@click.option(
    "-stpub", "--staking-public-key", type=str, default=None, help="Set Staking Public key for Cardano Shelley", show_default=True
)
@click.option(
    "-ct", "--cardano-type", type=str, default=Cardano.TYPES.SHELLEY_ICARUS, help="Select Cardano type", show_default=True
)
@click.option(
    "-at", "--address-type", type=str, default=None, help="Select Address type", show_default=True
)
@click.option(
    "-mo", "--mode", type=str, default=MODES.STANDARD, help="Select Mode of Electrum-V2", show_default=True
)
@click.option(
    "-mt", "--mnemonic-type", type=str, default=ELECTRUM_V2_MNEMONIC_TYPES.STANDARD, help="Select Mnemonic type of Electrum-V2", show_default=True
)
@click.option(
    "-cs", "--checksum", type=bool, default=False, help="Set Checksum for Monero", show_default=True
)
@click.option(
    "-se", "--semantic", type=str, default=None, help="Set Semantic for BIP141", show_default=True
)
@click.option(
    "-pi", "--payment-id", type=str, default=None, help="Set Payment ID for Monero", show_default=True
)
@click.option(
    "-ex", "--exclude", type=str, default="", help="Set Exclude keys from dumped", show_default=True
)
def cli_dump(**kwargs) -> None:  # cli_dumps(max_content_width=120)
    return dump(**kwargs)


@cli_main.command(
    "dumps", aliases=["ds"], options_metavar="[OPTIONS]", short_help="Select Dumps hdwallet keys"
)
@click.option(
    "-sym", "--symbol", type=str, default="BTC", help="Set Cryptocurrency ticker symbol", show_default=True
)
@click.option(
    "-h", "--hd", type=str, default="BIP44", help="Select HD", show_default=True
)
@click.option(
    "-n", "--network", type=str, default="mainnet", help="Select Network type", show_default=True
)
@click.option(
    "-ec", "--entropy-client", type=str, default="BIP39", help="Select Entropy client", show_default=True
)
@click.option(
    "-e", "--entropy", type=str, default=None, help="Set Master key from Entropy hex string", show_default=True
)
@click.option(
    "-mc", "--mnemonic-client", type=str, default="BIP39", help="Select Mnemonic client", show_default=True
)
@click.option(
    "-m", "--mnemonic", type=str, default=None, help="Set Master key from Mnemonic words", show_default=True
)
@click.option(
    "-l", "--language", type=str, default="english", help="Select Language for mnemonic", show_default=True
)
@click.option(
    "-sc", "--seed-client", type=str, default="BIP39", help="Select Seed client", show_default=True
)
@click.option(
    "-s", "--seed", type=str, default=None, help="Set Master key from Seed hex string", show_default=True
)
@click.option(
    "-pp", "--passphrase", type=str, default=None, help="Set Passphrase for mnemonic & seed", show_default=True
)
@click.option(
    "-xprv", "--xprivate-key", type=str, default=None, help="Set Master key from XPrivate key", show_default=True
)
@click.option(
    "-xpub", "--xpublic-key", type=str, default=None, help="Set Master key from XPublic key", show_default=True
)
@click.option(
    "-enc", "--encoded", type=bool, default=True, help="Set Encoded for XPrivate & XPublic keys", show_default=True
)
@click.option(
    "-st", "--strict", type=bool, default=False, help="Set Strict for Root XPrivate & XPublic keys", show_default=True
)
@click.option(
    "-pkt", "--public-key-type", type=str, default=PUBLIC_KEY_TYPES.COMPRESSED, help="Select Public key type", show_default=True
)
@click.option(
    "-d", "--derivation", type=str, default="BIP44", help="Select Derivation name", show_default=True
)
@click.option(
    "-ac", "--account", type=str, default="0", help="Set Account index for derivation", show_default=True
)
@click.option(
    "-ch", "--change", type=str, default="0", help="Set Change index for derivation", show_default=True
)
@click.option(
    "-ecc", "--ecc", type=str, default="0", help="Set ECC index for HDW derivation", show_default=True
)
@click.option(
    "-ro", "--role", type=str, default="0", help="Set Role index for CIP1852 derivation", show_default=True
)
@click.option(
    "-ad", "--address", type=str, default="0", help="Set Address index for derivation", show_default=True
)
@click.option(
    "-mi", "--minor", type=str, default="1", help="Set Minor index for Monero derivation", show_default=True
)
@click.option(
    "-ma", "--major", type=str, default="0", help="Set Major index for Monero derivation", show_default=True
)
@click.option(
    "-p", "--path", type=str, default=None, help="Set Path for derivation", show_default=True
)
@click.option(
    "-i", "--indexes", type=list, default=[], help="Set Indexes for derivation", show_default=True
)
@click.option(
    "-prv", "--private-key", type=str, default=None, help="Set Private key", show_default=True
)
@click.option(
    "-w", "--wif", type=str, default=None, help="Set Wallet Import Format (WIF)", show_default=True
)
@click.option(
    "-b38", "--bip38", type=bool, default=False, help="Is BIP38 Encrypted Wallet Import Format", show_default=True
)
@click.option(
    "-pub", "--public-key", type=str, default=None, help="Set Public key", show_default=True
)
@click.option(
    "-sprv", "--spend-private-key", type=str, default=None, help="Set Spend Private key for Monero", show_default=True
)
@click.option(
    "-vprv", "--view-private-key", type=str, default=None, help="Set View Private key for Monero", show_default=True
)
@click.option(
    "-spub", "--spend-public-key", type=str, default=None, help="Set Spend Public key for Monero", show_default=True
)
@click.option(
    "-stpub", "--staking-public-key", type=str, default=None, help="Set Staking Public key for Cardano Shelley", show_default=True
)
@click.option(
    "-ct", "--cardano-type", type=str, default=Cardano.TYPES.SHELLEY_ICARUS, help="Select Cardano type", show_default=True
)
@click.option(
    "-at", "--address-type", type=str, default=None, help="Select Address type", show_default=True
)
@click.option(
    "-mo", "--mode", type=str, default=MODES.STANDARD, help="Select Mode of Electrum-V2", show_default=True
)
@click.option(
    "-mt", "--mnemonic-type", type=str, default=ELECTRUM_V2_MNEMONIC_TYPES.STANDARD, help="Select Mnemonic type of Electrum-V2", show_default=True
)
@click.option(
    "-cs", "--checksum", type=bool, default=False, help="Set Checksum for Monero", show_default=True
)
@click.option(
    "-se", "--semantic", type=str, default=None, help="Set Semantic for BIP141", show_default=True
)
@click.option(
    "-pi", "--payment-id", type=str, default=None, help="Set Payment ID for Monero", show_default=True
)
@click.option(
    "-ex", "--exclude", type=str, default="", help="Set Exclude keys from dumped", show_default=True
)
@click.option(
    "-f", "--format", type=str, default="csv", help="Show dumps format type", show_default=True
)
@click.option(
    "-in", "--include", type=str, default=None, help="Set Include keys from dumped", show_default=True
)
@click.option(
    "-inh", "--include-header", is_flag=True, help="Set Include header from dumped", show_default=True
)
@click.option(
    "-de", "--delimiter", type=str, default=" ", help="Set Delimiter for CSV", show_default=True
)
def cli_dumps(**kwargs) -> None:  # cli_dumps(max_content_width=120)
    return dumps(**kwargs)


@cli_main.group(
    "list",
    aliases=["l"],
    cls=ClickAliasedGroup,
    options_metavar="[OPTIONS]",
    short_help="Select List for HDWallet information",
    invoke_without_command=True
)
def cli_list() -> None:
    pass


@cli_list.command(
    "cryptocurrencies",
    aliases=["c"],
    options_metavar="[OPTIONS]",
    short_help="List Available cryptocurrencies of HDWallet"
)
def cli_cryptocurrencies() -> None:
    return list_cryptocurrencies()


@cli_list.command(
    "languages",
    aliases=["l"],
    options_metavar="[OPTIONS]",
    short_help="List Languages of mnemonic words"
)
def cli_languages() -> None:
    return list_languages()


@cli_list.command(
    "strengths",
    aliases=["s"],
    options_metavar="[OPTIONS]",
    short_help="List Strengths of mnemonic words"
)
def cli_strengths() -> None:
    return list_strengths()



#!/usr/bin/env python3

# Copyright © 2020-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from ..slip44 import CoinTypes
from ..eccs import SLIP10Secp256k1ECC
from ..consts import (
    Info, Entropies, Mnemonics, Seeds, HDs, Addresses, Networks, XPrivateKeyVersions, XPublicKeyVersions
)
from .icryptocurrency import (
    ICryptocurrency, INetwork
)


class Mainnet(INetwork):

    NAME = "mainnet"
    PUBLIC_KEY_ADDRESS_PREFIX = 0xab3
    SCRIPT_ADDRESS_PREFIX = 0xab8
    XPRIVATE_KEY_VERSIONS = XPrivateKeyVersions({
        "P2PKH": 0x488ade3,
        "P2SH": 0x488ade3
    })
    XPUBLIC_KEY_VERSIONS = XPublicKeyVersions({
        "P2PKH": 0x488b21e,
        "P2SH": 0x488b21e
    })
    MESSAGE_PREFIX = "\x18LitecoinZ Signed Message:\n"
    WIF_PREFIX = 0x80
    

class LitecoinZ(ICryptocurrency):

    NAME = "LitecoinZ"
    SYMBOL = "LTZ"
    INFO = Info({
        "SOURCE_CODE": "https://github.com/litecoinz-project/litecoinz",
        "WHITEPAPER": "https://litecoinz.org/downloads/LITECOINZ-WHITE-PAPER.pdf",
        "WEBSITES": [
            "https://litecoinz.org"
        ]
    })
    ECC = SLIP10Secp256k1ECC
    COIN_TYPE = CoinTypes.LitecoinZ
    SUPPORT_BIP38 = True
    NETWORKS = Networks({
        "MAINNET": Mainnet
    })
    DEFAULT_NETWORK = NETWORKS.MAINNET
    ENTROPIES = Entropies({
        "BIP39"
    })
    MNEMONICS = Mnemonics({
        "BIP39"
    })
    SEEDS = Seeds({
        "BIP39"
    })
    HDS = HDs({
        "BIP32", "BIP44"
    })
    DEFAULT_HD = HDS.BIP44
    DEFAULT_PATH = f"m/44'/{COIN_TYPE}'/0'/0/0"
    ADDRESSES = Addresses({
        "P2PKH", "P2SH"
    })
    DEFAULT_ADDRESS = ADDRESSES.P2PKH
    SEMANTICS = [
        "p2pkh", "p2sh"
    ]
    DEFAULT_SEMANTIC = "p2pkh"



#!/usr/bin/env python3

# Copyright © 2020-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from ..point import SLIP10Ed25519Point


class SLIP10Ed25519Blake2bPoint(SLIP10Ed25519Point):

    @staticmethod
    def name() -> str:
        """
        Get the name of the ecc class.

        :return: The name of the ecc class.
        :rtype: str
        """

        return "SLIP10-Ed25519-Blake2b"



#!/usr/bin/env python3

# Copyright © 2020-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import (
    Tuple, Union, Optional, Dict
)

from ..libs.base58 import (
    encode_monero, decode_monero
)
from ..eccs import (
    IPublicKey, SLIP10Ed25519MoneroPublicKey, validate_and_get_public_key
)
from ..cryptocurrencies import Monero
from ..crypto import kekkak256
from ..exceptions import (
    Error, AddressError
)
from ..utils import (
    bytes_to_string, integer_to_bytes
)


class MoneroAddress:

    checksum_length: int = Monero.PARAMS.CHECKSUM_LENGTH
    payment_id_length: int = Monero.PARAMS.PAYMENT_ID_LENGTH
    networks: Dict[str, Dict[str, Dict[str, int]]] = {
        "mainnet": {
            "address_types": {
                "standard": Monero.NETWORKS.MAINNET.STANDARD,
                "integrated": Monero.NETWORKS.MAINNET.INTEGRATED,
                "sub-address": Monero.NETWORKS.MAINNET.SUB_ADDRESS
            }
        },
        "stagenet": {
            "address_types": {
                "standard": Monero.NETWORKS.STAGENET.STANDARD,
                "integrated": Monero.NETWORKS.STAGENET.INTEGRATED,
                "sub-address": Monero.NETWORKS.STAGENET.SUB_ADDRESS
            }
        },
        "testnet": {
            "address_types": {
                "standard": Monero.NETWORKS.TESTNET.STANDARD,
                "integrated": Monero.NETWORKS.TESTNET.INTEGRATED,
                "sub-address": Monero.NETWORKS.TESTNET.SUB_ADDRESS
            }
        }
    }

    @staticmethod
    def name() -> str:
        """
        Returns the name of the cryptocurrency, which is "Monero".

        :return: The name of the cryptocurrency.
        :rtype: str
        """

        return "Monero"

    @classmethod
    def compute_checksum(cls, public_key: bytes) -> bytes:
        return kekkak256(public_key)[:cls.checksum_length]

    @classmethod
    def encode(
        cls,
        spend_public_key: Union[bytes, str, IPublicKey],
        view_public_key: Union[bytes, str, IPublicKey],
        network: str = "mainnet",
        address_type: str = Monero.ADDRESS_TYPES.STANDARD,
        payment_id: Optional[bytes] = None
    ) -> str:
        """
        Encodes the given spend and view public keys into a Monero address.

        :param spend_public_key: The spend public key.
        :type spend_public_key: Union[bytes, str, IPublicKey]
        :param view_public_key: The view public key.
        :type view_public_key: Union[bytes, str, IPublicKey]
        :param network: The network (default is "mainnet").
        :type network: str
        :param address_type: The address type (default is Monero.ADDRESS_TYPES.STANDARD).
        :type address_type: str
        :param payment_id: Optional payment ID bytes.
        :type payment_id: Optional[bytes]

        :return: The encoded Monero address.
        :rtype: str
        """

        spend_public_key: IPublicKey = validate_and_get_public_key(
            public_key=spend_public_key, public_key_cls=SLIP10Ed25519MoneroPublicKey
        )
        view_public_key: IPublicKey = validate_and_get_public_key(
            public_key=view_public_key, public_key_cls=SLIP10Ed25519MoneroPublicKey
        )

        if payment_id is not None and len(payment_id) != cls.payment_id_length:
            raise Error("Invalid payment ID length")

        version: bytes = integer_to_bytes(
            cls.networks[network]["address_types"][address_type]
        )
        payload: bytes = (
            version + spend_public_key.raw_compressed() +
            view_public_key.raw_compressed() +
            (b"" if payment_id is None else payment_id)
        )

        return encode_monero(payload + cls.compute_checksum(payload))

    @classmethod
    def decode(
        cls,
        address: str,
        network: str = "mainnet",
        address_type: str = Monero.ADDRESS_TYPES.STANDARD,
        payment_id: Optional[bytes] = None
    ) -> Tuple[str, str]:
        """
        Decodes a Monero address into spend and view public keys.

        :param address: The Monero address to decode.
        :type address: str
        :param network: The network (default is "mainnet").
        :type network: str
        :param address_type: The address type (default is Monero.ADDRESS_TYPES.STANDARD).
        :type address_type: str
        :param payment_id: Optional payment ID bytes.
        :type payment_id: Optional[bytes]

        :return: A tuple containing the spend public key and view public key.
        :rtype: Tuple[str, str]
        """

        address_decode: bytes = decode_monero(address)

        checksum: bytes = address_decode[-1 * cls.checksum_length:]
        payload_with_prefix: bytes = address_decode[:-1 * cls.checksum_length]

        checksum_got: bytes = cls.compute_checksum(payload_with_prefix)
        if checksum != checksum_got:
            raise AddressError(
                "Invalid checksum", expected=bytes_to_string(checksum), got=bytes_to_string(checksum_got)
            )

        version: bytes = integer_to_bytes(
            cls.networks[network]["address_types"][address_type]
        )
        version_got = payload_with_prefix[:len(version)]
        if version != version_got:
            raise AddressError(
                "Invalid version", expected=version, got=version_got
            )

        payload_without_prefix: bytes = payload_with_prefix[len(version):]

        expected_length: int = SLIP10Ed25519MoneroPublicKey.compressed_length() * 2
        try:
            if len(payload_without_prefix) != expected_length:
                raise AddressError(
                    "Invalid length", expected=expected_length, got=len(payload_without_prefix)
                )
        except ValueError as ex:
            if len(payload_without_prefix) != expected_length + cls.payment_id_length:
                raise AddressError(
                    "Invalid length",
                    expected=(expected_length + cls.payment_id_length),
                    got=len(payload_without_prefix)
                )

            if payment_id is None or len(payment_id) != cls.payment_id_length:
                raise Error("Invalid payment ID")

            payment_id_got_bytes = payload_without_prefix[-cls.payment_id_length:]
            if payment_id != payment_id_got_bytes:
                raise Error(
                    "Invalid payment ID",
                    expected=bytes_to_string(payment_id_got_bytes),
                    got=bytes_to_string(payment_id_got_bytes)
                )

        length: int = SLIP10Ed25519MoneroPublicKey.compressed_length()

        spend_public_key: bytes = payload_without_prefix[:length]
        if not SLIP10Ed25519MoneroPublicKey.is_valid_bytes(spend_public_key):
            raise Error(f"Invalid {SLIP10Ed25519MoneroPublicKey.name()} public key {bytes_to_string(spend_public_key)}")

        view_public_key: bytes = payload_without_prefix[length:(length * 2)]
        if not SLIP10Ed25519MoneroPublicKey.is_valid_bytes(view_public_key):
            raise Error(f"Invalid {SLIP10Ed25519MoneroPublicKey.name()} public key {bytes_to_string(view_public_key)}")

        return bytes_to_string(spend_public_key), bytes_to_string(view_public_key)



#!/usr/bin/env python3

# Copyright © 2020-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from ..slip44 import CoinTypes
from ..eccs import SLIP10Secp256k1ECC
from ..consts import (
    Info, Entropies, Mnemonics, Seeds, HDs, Addresses, Networks, XPrivateKeyVersions, XPublicKeyVersions
)
from .icryptocurrency import (
    ICryptocurrency, INetwork
)


class Mainnet(INetwork):

    NAME = "mainnet"
    PUBLIC_KEY_ADDRESS_PREFIX = 0x19
    SCRIPT_ADDRESS_PREFIX = 0x55
    XPRIVATE_KEY_VERSIONS = XPrivateKeyVersions({
        "P2PKH": 0x2cfbf60,
        "P2SH": 0x2cfbf60
    })
    XPUBLIC_KEY_VERSIONS = XPublicKeyVersions({
        "P2PKH": 0x2cfbede,
        "P2SH": 0x2cfbede
    })
    MESSAGE_PREFIX = "\x18BlackCoin Signed Message:\n"
    WIF_PREFIX = 0x99


class Blackcoin(ICryptocurrency):

    NAME = "Blackcoin"
    SYMBOL = "BLK"
    INFO = Info({
        "SOURCE_CODE": "https://github.com/coinblack",
        "WHITEPAPER": "https://blackcoin.org/blackcoin-pos-protocol-v2-whitepaper.pdf",
        "WEBSITES": [
            "https://blackcoin.org",
            "https://blackcoinmore.org"
        ]
    })
    ECC = SLIP10Secp256k1ECC
    COIN_TYPE = CoinTypes.Blackcoin
    SUPPORT_BIP38 = True
    NETWORKS = Networks({
        "MAINNET": Mainnet
    })
    DEFAULT_NETWORK = NETWORKS.MAINNET
    ENTROPIES = Entropies({
        "BIP39"
    })
    MNEMONICS = Mnemonics({
        "BIP39"
    })
    SEEDS = Seeds({
        "BIP39"
    })
    HDS = HDs({
        "BIP32", "BIP44"
    })
    DEFAULT_HD = HDS.BIP44
    DEFAULT_PATH = f"m/44'/{COIN_TYPE}'/0'/0/0"
    ADDRESSES = Addresses({
        "P2PKH", "P2SH"
    })
    DEFAULT_ADDRESS = ADDRESSES.P2PKH
    SEMANTICS = [
        "p2pkh", "p2sh"
    ]
    DEFAULT_SEMANTIC = "p2pkh"



#!/usr/bin/env python3

# Copyright © 2020-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import (
    Any, Union
)

from ..libs.base58 import (
    ensure_string, encode, decode
)
from ..eccs import (
    IPublicKey, SLIP10Ed25519PublicKey, validate_and_get_public_key
)
from ..cryptocurrencies import Solana
from ..utils import bytes_to_string
from .iaddress import IAddress


class SolanaAddress(IAddress):

    alphabet: str = Solana.PARAMS.ALPHABET

    @staticmethod
    def name() -> str:
        """
        Return the name of the cryptocurrency, which is "Solana".

        :return: The name of the cryptocurrency.
        :rtype: str
        """

        return "Solana"

    @classmethod
    def encode(cls, public_key: Union[bytes, str, IPublicKey], **kwargs: Any) -> str:
        """
        Encode a public key into a Solana address.

        :param public_key: The public key to encode. It can be a bytes object, a string, or an IPublicKey instance.
        :type public_key: Union[bytes, str, IPublicKey]
        :param kwargs: Additional keyword arguments.
        :type kwargs: Any

        :return: The encoded Solana address as a string.
        :rtype: str
        """

        public_key: IPublicKey = validate_and_get_public_key(
            public_key=public_key, public_key_cls=SLIP10Ed25519PublicKey
        )

        return ensure_string(encode(public_key.raw_compressed()[1:]))

    @classmethod
    def decode(cls, address: str, **kwargs: Any) -> str:
        """
        Decode a Solana address back into its corresponding public key.

        :param address: The Solana address to decode.
        :type address: str

        :param kwargs: Additional keyword arguments.
        :type kwargs: Any

        :return: The decoded public key as a string.
        :rtype: str
        """

        public_key: bytes = decode(address)

        expected_length = SLIP10Ed25519PublicKey.compressed_length() - 1
        if len(public_key) != expected_length:
            raise ValueError(f"Invalid length (expected: {expected_length}, got: {len(public_key)})")

        if not SLIP10Ed25519PublicKey.is_valid_bytes(public_key):
            raise ValueError(f"Invalid {SLIP10Ed25519PublicKey.name()} public key {bytes_to_string(public_key)}")

        return bytes_to_string(public_key)



#!/usr/bin/env python3

# Copyright © 2020-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import (
    Any, Union
)

from ..libs.base58 import (
    ensure_string, check_encode
)
from ..consts import PUBLIC_KEY_TYPES
from ..eccs import (
    IPublicKey, SLIP10Secp256k1PublicKey, validate_and_get_public_key
)
from ..crypto import (
    hash160, sha256
)
from ..utils import (
    get_bytes, integer_to_bytes, bytes_to_string
)
from .p2sh import P2SHAddress


class P2WSHInP2SHAddress(P2SHAddress):

    @staticmethod
    def name() -> str:
        """
        Returns the name "P2WSH-In-P2SH".

        :return: The name of the address type.
        :rtype: str
        """

        return "P2WSH-In-P2SH"

    @classmethod
    def encode(cls, public_key: Union[bytes, str, IPublicKey], **kwargs: Any) -> str:
        """
        Encode a public key into a P2WSH address.

        :param public_key: The public key to encode.
        :type public_key: Union[bytes, str, IPublicKey]
        :param kwargs: Additional keyword arguments.
            - script_address_prefix: Script address prefix (optional).
            - public_key_type: Type of public key (optional).
            - alphabet: Custom alphabet for encoding (optional).
        :type kwargs: Any

        :return: The encoded P2WSH address.
        :rtype: str
        """

        script_address_prefix: bytes = integer_to_bytes(
            kwargs.get("script_address_prefix", cls.script_address_prefix)
        )
        public_key: IPublicKey = validate_and_get_public_key(
            public_key=public_key, public_key_cls=SLIP10Secp256k1PublicKey
        )
        public_key_bytes: bytes = (
            public_key.raw_compressed()
            if kwargs.get("public_key_type", PUBLIC_KEY_TYPES.COMPRESSED) == PUBLIC_KEY_TYPES.COMPRESSED else
            public_key.raw_uncompressed()
        )
        script_hash: bytes = hash160(get_bytes(
            "0020" + bytes_to_string(sha256(get_bytes(
                "5121" + bytes_to_string(public_key_bytes) + "51ae"
            )))
        ))

        return ensure_string(check_encode(
            (script_address_prefix + script_hash), alphabet=kwargs.get(
                "alphabet", cls.alphabet
            )
        ))



#!/usr/bin/env python3

# Copyright © 2020-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import (
    Any, Union
)

from ..libs.base32 import (
    encode_no_padding, decode
)
from ..eccs import (
    IPublicKey, SLIP10Ed25519PublicKey, validate_and_get_public_key
)
from ..cryptocurrencies import Stellar
from ..crypto import xmodem_crc
from ..utils import (
    get_bytes, bytes_reverse, integer_to_bytes, bytes_to_string
)
from ..exceptions import AddressError
from .iaddress import IAddress


class StellarAddress(IAddress):

    checksum_length: int = Stellar.PARAMS.CHECKSUM_LENGTH
    address_types: dict = {
        "private_key": Stellar.PARAMS.ADDRESS_TYPES.PRIVATE_KEY,
        "public_key": Stellar.PARAMS.ADDRESS_TYPES.PUBLIC_KEY
    }

    @staticmethod
    def name() -> str:
        """
        Return the name of the cryptocurrency, which is "Stellar".

        :return: The name of the cryptocurrency.
        :rtype: str
        """

        return "Stellar"

    @staticmethod
    def compute_checksum(public_key: bytes) -> bytes:
        return bytes_reverse(xmodem_crc(public_key))

    @classmethod
    def encode(cls, public_key: Union[bytes, str, IPublicKey], **kwargs: Any) -> str:
        """
        Encode a public key into a Stellar address.

        :param public_key: The public key to encode.
        :type public_key: Union[bytes, str, IPublicKey]

        :param kwargs: Additional keyword arguments.
            - address_type: Type of the Stellar address (optional).
        :type kwargs: Any

        :return: The encoded Stellar address.
        :rtype: str
        """

        if not kwargs.get("address_type"):
            address_type: int = cls.address_types[Stellar.DEFAULT_ADDRESS_TYPE]
        else:
            if kwargs.get("address_type") not in Stellar.ADDRESS_TYPES.get_address_types():
                raise AddressError(
                    f"Invalid {cls.name()} address type",
                    expected=Stellar.ADDRESS_TYPES.get_address_types(),
                    got=kwargs.get("address_type")
                )
            address_type: int = cls.address_types[kwargs.get("address_type")]

        public_key: IPublicKey = validate_and_get_public_key(
            public_key=public_key, public_key_cls=SLIP10Ed25519PublicKey
        )
        payload: bytes = integer_to_bytes(address_type) + public_key.raw_compressed()[1:]
        checksum: bytes = cls.compute_checksum(payload)

        return encode_no_padding((payload + checksum).hex())

    @classmethod
    def decode(cls, address: str, **kwargs: Any) -> str:
        """
        Decode a Stellar address back into its public key representation.

        :param address: The Stellar address to decode.
        :type address: str

        :param kwargs: Additional keyword arguments.
            - address_type: Type of the Stellar address (optional).
        :type kwargs: Any

        :return: The decoded public key as a string.
        :rtype: str
        """

        if not kwargs.get("address_type"):
            address_type: int = cls.address_types[Stellar.DEFAULT_ADDRESS_TYPE]
        else:
            if kwargs.get("address_type") not in Stellar.ADDRESS_TYPES.get_address_types():
                raise AddressError(
                    f"Invalid {cls.name()} address type",
                    expected=Stellar.ADDRESS_TYPES.get_address_types(),
                    got=kwargs.get("address_type")
                )
            address_type: int = cls.address_types[kwargs.get("address_type")]

        address_decode: bytes = get_bytes(decode(address))
        expected_length: int = (
            SLIP10Ed25519PublicKey.compressed_length() + cls.checksum_length
        )
        if len(address_decode) != expected_length:
            raise ValueError(f"Invalid length (expected: {expected_length}, got: {len(address_decode)})")

        checksum: bytes = address_decode[-1 * cls.checksum_length:]
        payload: bytes = address_decode[:-1 * cls.checksum_length]

        address_type_got: int = payload[0]
        if address_type != address_type_got:
            raise ValueError(f"Invalid address type (expected: {address_type}, got: {address_type_got})")

        checksum_got: bytes = cls.compute_checksum(payload)
        if checksum != checksum_got:
            raise ValueError(f"Invalid checksum (expected: {bytes_to_string(checksum)}, got: {bytes_to_string(checksum_got)})")

        public_key: bytes = payload[1:]
        if not SLIP10Ed25519PublicKey.is_valid_bytes(public_key):
            raise ValueError(f"Invalid {SLIP10Ed25519PublicKey.name()} public key {bytes_to_string(public_key)}")

        return bytes_to_string(public_key)



#!/usr/bin/env python3

# Copyright © 2020-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import (
    Any, Union
)

from ..libs.base58 import (
    ensure_string, check_encode, check_decode
)
from ..eccs import (
    IPublicKey, SLIP10Secp256k1PublicKey, validate_and_get_public_key
)
from ..cryptocurrencies import Tron
from ..crypto import kekkak256
from ..utils import (
    integer_to_bytes, bytes_to_string
)
from .iaddress import IAddress


class TronAddress(IAddress):

    public_key_address_prefix: bytes = integer_to_bytes(
        Tron.NETWORKS.MAINNET.PUBLIC_KEY_ADDRESS_PREFIX
    )
    alphabet: str = Tron.PARAMS.ALPHABET

    @staticmethod
    def name() -> str:
        """
        Return the name associated with the Tron blockchain protocol.

        :return: The name "Tron".
        :rtype: str
        """

        return "Tron"

    @classmethod
    def encode(cls, public_key: Union[bytes, str, IPublicKey], **kwargs: Any) -> str:
        """
        Encode a public key into a Tron blockchain address.

        :param public_key: The public key to encode, which can be bytes, str, or IPublicKey.
        :type public_key: Union[bytes, str, IPublicKey]
        :param kwargs: Additional keyword arguments.
        :type kwargs: Any

        :return: Encoded Tron blockchain address.
        :rtype: str
        """

        public_key: IPublicKey = validate_and_get_public_key(
            public_key=public_key, public_key_cls=SLIP10Secp256k1PublicKey
        )
        
        address: str = bytes_to_string(
            kekkak256(public_key.raw_uncompressed()[1:])
        )[24:]

        return ensure_string(check_encode(
            (cls.public_key_address_prefix + bytearray.fromhex(address)), alphabet=kwargs.get(
                "alphabet", cls.alphabet
            )
        ))

    @classmethod
    def decode(cls, address: str, **kwargs: Any) -> str:
        """
        Decode a Tron blockchain address back to its public key representation.

        :param address: The Tron blockchain address to decode.
        :type address: str
        :param kwargs: Additional keyword arguments.
        :type kwargs: Any

        :return: Decoded public key.
        :rtype: str
        """
        
        address_decode: bytes = check_decode(
            address, alphabet=kwargs.get(
                "alphabet", cls.alphabet
            )
        )

        expected_length: int = 20 + len(cls.public_key_address_prefix)
        if len(address_decode) != expected_length:
            raise ValueError(f"Invalid length (expected: {expected_length}, got: {len(address_decode)})")

        prefix_got: bytes = address_decode[:len(cls.public_key_address_prefix)]
        if cls.public_key_address_prefix != prefix_got:
            raise ValueError(f"Invalid prefix (expected: {cls.public_key_address_prefix}, got: {prefix_got})")

        return bytes_to_string(address_decode[len(cls.public_key_address_prefix):])



#!/usr/bin/env python3

# Copyright © 2020-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import (
    Optional, Union
)

import unicodedata

from ...crypto import pbkdf2_hmac_sha512
from ...exceptions import MnemonicError
from ...utils import bytes_to_string
from ...mnemonics import (
    IMnemonic, ElectrumV2Mnemonic, ELECTRUM_V2_MNEMONIC_TYPES
)
from ..iseed import ISeed


class ElectrumV2Seed(ISeed):
    """
    This class generates a root extended private key from a given seed using the
    Electrum-V2 standard. The Electrum-V2 standard defines a method for generating mnemonic
    phrases and converting them into a binary seed used for hierarchical
    deterministic wallets.

    .. note::
        This class inherits from the ``ISeed`` class, thereby ensuring that all functions are accessible.
    """

    seed_salt_modifier: str = "electrum"
    seed_pbkdf2_rounds: int = 2048

    length = 128

    @classmethod
    def name(cls) -> str:
        """
        Get the name of the seeds class.

        :return: The name of the seeds class.
        :rtype: str
        """

        return "Electrum-V2"

    @classmethod
    def from_mnemonic(
        cls,
        mnemonic: Union[str, IMnemonic],
        passphrase: Optional[str] = None,
        mnemonic_type=ELECTRUM_V2_MNEMONIC_TYPES.STANDARD
    ) -> str:
        """
        Converts an Electrum V2 mnemonic phrase to its corresponding seed.

        :param mnemonic: The mnemonic phrase to be decoded. Can be a string or an instance of `IMnemonic`.
        :type mnemonic: Union[str, IMnemonic]
        :param passphrase: An optional passphrase to strengthen the security of the seed. Defaults to None.
        :type passphrase: Optional[str]
        :param mnemonic_type: The type of Electrum V2 mnemonic, defaults to STANDARD.
        :type mnemonic_type: str

        :return: The derived seed as a string.
        :rtype: str
        """

        mnemonic = (
            mnemonic.mnemonic() if isinstance(mnemonic, IMnemonic) else mnemonic
        )
        if not ElectrumV2Mnemonic.is_valid(mnemonic=mnemonic, mnemonic_type=mnemonic_type):
            raise MnemonicError(f"Invalid {cls.name()} mnemonic words")

        salt: str = unicodedata.normalize("NFKD", (
            (cls.seed_salt_modifier + passphrase) if passphrase else cls.seed_salt_modifier
        ))
        return bytes_to_string(pbkdf2_hmac_sha512(
            password=mnemonic, salt=salt, iteration_num=cls.seed_pbkdf2_rounds
        ))



#!/usr/bin/env python3

# Copyright © 2020-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import (
    Any, Union
)

from ..libs.bech32 import (
    bech32_encode, bech32_decode
)
from ..eccs import (
    IPublicKey, SLIP10Ed25519PublicKey, validate_and_get_public_key
)
from ..cryptocurrencies import MultiversX
from ..utils import bytes_to_string
from .iaddress import IAddress


class MultiversXAddress(IAddress):

    hrp: str = MultiversX.NETWORKS.MAINNET.HRP

    @staticmethod
    def name() -> str:
        """
        Returns the name of the cryptocurrency, which is "MultiversX".

        :return: The name of the cryptocurrency.
        :rtype: str
        """

        return "MultiversX"

    @classmethod
    def encode(cls, public_key: Union[bytes, str, IPublicKey], **kwargs: Any) -> str:
        """
        Encode the public key into a string format using Bech32 encoding.

        :param public_key: The public key to encode, can be bytes, str, or IPublicKey.
        :type public_key: Union[bytes, str, IPublicKey]
        :param kwargs: Additional keyword arguments.
            - hrp: Human-readable part (optional).
        :type kwargs: Any

        :return: The encoded public key as a string.
        :rtype: str
        """

        public_key: IPublicKey = validate_and_get_public_key(
            public_key=public_key, public_key_cls=SLIP10Ed25519PublicKey
        )
        return bech32_encode(
            kwargs.get("hrp", cls.hrp), public_key.raw_compressed()[1:]
        )

    @classmethod
    def decode(cls, address: str, **kwargs: Any) -> str:
        """
        Decode an address string using Bech32 decoding.

        :param address: The address string to decode.
        :type address: str
        :param kwargs: Additional keyword arguments.
            - hrp: Human-readable part (optional).
        :type kwargs: Any

        :return: The decoded address as a string.
        :rtype: str
        """

        hrp, address_decode = bech32_decode(
            kwargs.get("hrp", cls.hrp), address
        )
        return bytes_to_string(address_decode)



#!/usr/bin/env python3

# Copyright © 2020-2024, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import (
    Optional, List, Tuple, Union
)

from ..utils import normalize_derivation


class IDerivation:

    _path: str = "m/"
    _indexes: List[int] = []
    _derivations: List[Tuple[int, bool]] = { }

    def __init__(
        self, path: Optional[str] = None, indexes: Optional[List[int]] = None, **kwargs
    ) -> None:
        """
        Initializes an object for iderivation.

        :param path: Optional derivation path string.
        :type path: Optional[str]
        :param indexes: Optional list of derivation indexes.
        :type indexes: Optional[List[int]]

        :return: No return
        :rtype: NoneType
        """

        self._path, self._indexes, self._derivations = normalize_derivation(
            path=path, indexes=indexes
        )

    def __str__(self) -> str:
        """
        Return the string representation of the derivation path.

        :return: The derivation path as a string.
        :rtype: str
        """

        return self._path

    @classmethod
    def name(cls) -> str:
        """
        Get the name of the derivation class.

        :return: The name of the derivation class.
        :rtype: str
        """

        pass

    def clean(self) -> "IDerivation":
        """
        Reset the object's attributes to their initial states or defaults.

        :return: The updated `IDerivation` object itself after cleaning.
        :rtype: IDerivation
        """

        pass

    def path(self) -> str:
        """
        Retrieves the path associated with the current instance.

        :return: The derivation as a string.
        :rtype: str
        """

        return self._path

    def indexes(self) -> List[int]:
        """
        Retrieve the list of indexes in the derivation path.

        :return: A list of integer indexes used in the derivation path.
        :rtype: List[int]
        """

        return self._indexes

    def derivations(self) -> List[Tuple[int, bool]]:
        """
        Retrieve the list of derivations in the derivation path.

        :return: A list of tuples where each tuple contains an index and a boolean indicating whether the index is hardened.
        :rtype: List[Tuple[int, bool]]
        """

        return self._derivations

    def depth(self) -> int:
        """
        Retrieve the depth of the derivation path.

        :return: The number of derivation levels in the path.
        :rtype: int
        """

        return len(self._derivations)

    def purpose(self) -> int:
        """
        Retrieve the purpose value from the object's `_purpose` attribute.

        This method returns the first element of the `_purpose` attribute,
        which is assumed to be a list or similar iterable containing integers.

        :return: The purpose value stored as the first element of `_purpose`.
        :rtype: int
        """

        pass

    def coin_type(self) -> int:
        """
        Retrieve the coin type value from the object's `_coin_type` attribute.

        This method returns the value of the `_coin_type` attribute,
        which is expected to be an integer representing the type of coin.

        :return: The coin type value stored in `_coin_type`.
        :rtype: int
        """

        pass

    def account(self) -> int:
        """
        Retrieve the account value from the object's `_account` attribute.

        This method returns the value of the `_account` attribute,
        which is expected to be an integer representing the account identifier.

        :return: The account value stored in `_account`.
        :rtype: int
        """

        pass

    def change(self) -> Union[int, str]:
        """
        Retrieve the change value from the object's `_change` attribute.

        This method returns the value of the `_change` attribute,
        which can be an integer or a string, representing the change status or type.

        :return: The change value stored in `_change`.
        :rtype: Union[int, str]
        """

        pass

    def role(self) -> str:
        """
        Retrieve the role value from the object's `_role` attribute.

        This method returns the value of the `_role` attribute,
        which is expected to be a string representing the role of the object.

        :return: The role value stored in `_role`.
        :rtype: str
        """

        pass

    def address(self) -> int:
        """
        Retrieve the address value from the object's `_address` attribute.

        This method returns the value of the `_address` attribute,
        which is expected to be an integer representing the address.

        :return: The address value stored in `_address`.
        :rtype: int
        """

        pass

    def minor(self) -> int:
        """
        Retrieve the minor version value from the object's `_minor` attribute.

        This method returns the value of the `_minor` attribute,
        which is expected to be an integer representing the minor version number.

        :return: The minor version value stored in `_minor`.
        :rtype: int
        """

        pass

    def major(self) -> int:
        """
        Retrieve the major version value from the object's `_major` attribute.

        This method returns the value of the `_major` attribute,
        which is expected to be an integer representing the major version number.

        :return: The major version value stored in `_major`.
        :rtype: int
        """

        pass



#!/usr/bin/env python3

# Copyright © 2020-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import Any
from ecdsa import SigningKey
from ecdsa import (
    curves, keys
)

import coincurve

from ....consts import SLIP10_SECP256K1_CONST
from ...iecc import (
    IPublicKey, IPrivateKey
)
from .public_key import (
    SLIP10Secp256k1PublicKeyCoincurve, SLIP10Secp256k1PublicKeyECDSA
)


class SLIP10Secp256k1PrivateKeyCoincurve(IPrivateKey):

    signing_key: coincurve.PrivateKey

    def __init__(self, private_key: coincurve.PrivateKey) -> None:
        """
        Initializes an instance of SLIP10 Secp256k1 private key.

        :param private_key: The coincurve PrivateKey object representing the private key.
        :type private_key: coincurve.PrivateKey
        """

        self.signing_key = private_key

    @staticmethod
    def name() -> str:
        """
        Get the name of the ecc class.

        :return: The name of the ecc class.
        :rtype: str
        """

        return "SLIP10-Secp256k1"

    @classmethod
    def from_bytes(cls, private_key: bytes) -> IPrivateKey:
        """
        Creates an instance of SLIP10 Secp256k1 private key from bytes.

        :param private_key: The private key bytes.
        :type private_key: bytes

        :return: An instance implementing the IPrivateKey interface.
        :rtype: IPrivateKey
        """

        if len(private_key) != cls.length():
            raise ValueError("Invalid private key bytes")

        try:
            return cls(coincurve.PrivateKey(private_key))
        except ValueError as ex:
            raise ValueError("Invalid private key bytes") from ex

    @staticmethod
    def length() -> int:
        """
        Returns the length of the SLIP10 Secp256k1 private key in bytes.

        :return: The length of the private key in bytes.
        :rtype: int
        """

        return SLIP10_SECP256K1_CONST.PRIVATE_KEY_BYTE_LENGTH

    def underlying_object(self) -> Any:
        """
        Retrieves the underlying signing key object.

        :return: The underlying signing key object.
        :rtype: Any
        """

        return self.signing_key

    def raw(self) -> bytes:
        """
        Retrieves the raw secret bytes of the signing key.

        :return: The raw secret bytes of the signing key.
        :rtype: bytes
        """

        return self.signing_key.secret

    def public_key(self) -> IPublicKey:
        """
        Retrieve the public key associated with this private key instance.

        :return: The public key object.
        :rtype: IPublicKey
        """

        return SLIP10Secp256k1PublicKeyCoincurve(self.signing_key.public_key)


class SLIP10Secp256k1PrivateKeyECDSA(IPrivateKey):

    signing_key: SigningKey

    def __init__(self, signing_key: SigningKey) -> None:
        """
        Initializes an instance with a signing key.

        :param signing_key: The signing key to be used for cryptographic operations.
        :type signing_key: SigningKey
        """

        self.signing_key = signing_key

    @staticmethod
    def name() -> str:
        """
        Returns the name of the SLIP10-Secp256k1 algorithm.

        :return: The name of the algorithm.
        :rtype: str
        """

        return "SLIP10-Secp256k1"

    @classmethod
    def from_bytes(cls, key_bytes: bytes) -> IPrivateKey:
        """
        Creates a private key instance from the given bytes.

        :param key_bytes: The bytes representing the private key.
        :type key_bytes: bytes

        :return: An instance of IPrivateKey corresponding to the given bytes.
        :rtype: IPrivateKey
        """

        try:
            return cls(
                SigningKey.from_string(
                    key_bytes, curve=curves.SECP256k1
                )
            )
        except keys.MalformedPointError as ex:
            raise ValueError("Invalid private key bytes") from ex

    @staticmethod
    def length() -> int:
        """
        Returns the length of the private key in bytes for SLIP10 SECP256k1 curve.

        :return: Length of the private key in bytes.
        :rtype: int

        """

        return SLIP10_SECP256K1_CONST.PRIVATE_KEY_BYTE_LENGTH

    def underlying_object(self) -> Any:
        """
        Return the underlying signing key object.

        :return: The underlying signing key object.
        :rtype: Any
        """

        return self.signing_key

    def raw(self) -> bytes:
        """
        Return the raw bytes representation of the private key.

        :return: The raw bytes of the private key.
        :rtype: bytes
        """

        return self.signing_key.to_string()

    def public_key(self) -> IPublicKey:
        """
        Retrieve the public key associated with this private key instance.

        :return: The public key object.
        :rtype: IPublicKey
        """

        return SLIP10Secp256k1PublicKeyECDSA(self.signing_key.get_verifying_key())


if SLIP10_SECP256K1_CONST.USE == "coincurve":
    SLIP10Secp256k1PrivateKey = SLIP10Secp256k1PrivateKeyCoincurve
elif SLIP10_SECP256K1_CONST.USE == "ecdsa":
    SLIP10Secp256k1PrivateKey = SLIP10Secp256k1PrivateKeyECDSA
else:
    raise Exception(
        f"Invalid SLIP10-Secp256k1 use, (expected: 'coincurve' or 'ecdsa', got: '{SLIP10_SECP256K1_CONST.USE}')"
    )



#!/usr/bin/env python3

# Copyright © 2020-2024, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from random import choice
from typing import (
    List, Tuple, AnyStr, Optional, Union, Literal
)

import binascii
import string
import re

from .exceptions import DerivationError


def generate_passphrase(length: int = 32) -> str:
    """
    Generate a random passphrase.

    :param length: The length of the passphrase (default is 32).
    :type length: int

    :return: A randomly generated passphrase consisting of ASCII letters and digits.
    :rtype: str
    """

    return "".join(choice(string.ascii_letters + string.digits) for _ in range(length))


def get_hmac(ecc_name: str) -> bytes:
    """
    Return the HMAC seed for the specified elliptic curve algorithm.

    :param ecc_name: The name of the elliptic curve algorithm.
    :type ecc_name: str

    :return: The HMAC seed corresponding to the elliptic curve algorithm.
    :rtype: bytes
    """

    if ecc_name in [
        "Kholaw-Ed25519", "SLIP10-Ed25519", "SLIP10-Ed25519-Blake2b", "SLIP10-Ed25519-Monero"
    ]:
        return b"ed25519 seed"
    elif ecc_name == "SLIP10-Nist256p1":
        return b"Nist256p1 seed"
    elif ecc_name == "SLIP10-Secp256k1":
        return b"Bitcoin seed"


def exclude_keys(nested: dict, keys: set) -> dict:
    """
    Recursively exclude keys from a nested dictionary based on a set of keys.

    :param nested: The nested dictionary from which keys are to be excluded.
    :type nested: dict
    :param keys: A set of keys to exclude from the dictionary. Keys are checked after converting '-' to '_'.
    :type keys: set

    :return: A new dictionary with excluded keys.
    :rtype: dict
    """

    new: dict = { }
    for _key, _value in nested.items():
        if isinstance(_value, dict):
            new[_key] = exclude_keys(_value, keys)
        elif _key not in [key.replace("-", "_") if isinstance(key, str) else key for key in keys]:
            new[_key] = _value
    return new


def path_to_indexes(path: str) -> List[int]:
    """
    Convert a derivation path string into a list of indexes.

    :param path: The derivation path string, e.g., "m/0'/1/2'/2".
                 Should follow the format "m/0'/0" or similar.
    :type path: str

    :return: A list of indexes derived from the path, where hardened indexes
             have the highest bit set (0x80000000).
    :rtype: List[int]
        """

    if path in ["m", "m/"]:
        return []
    elif path[0:2] != "m/":
        raise DerivationError(
            f"Bad path format", expected="like this type of path \"m/0'/0\"", got=path
        )

    indexes: List[int] = []
    for index in path.lstrip("m/").split("/"):
        indexes.append((int(index[:-1]) + 0x80000000) if "'" in index else int(index))
    return indexes


def indexes_to_path(indexes: List[int]) -> str:
    """
    Convert a list of indexes into a derivation path string.

    :param indexes: A list of indexes, where hardened indexes have the highest bit set (0x80000000).
    :type indexes: List[int]

    :return: The derivation path string generated from the list of indexes.
    :rtype: str
    """

    path: str = "m"
    for index in indexes:
        path += f"/{index - 0x80000000}'" if index & 0x80000000 else f"/{index}"
    return path


def normalize_index(
    index: Union[str, int, Tuple[int, int]], hardened: bool = False
) -> Union[Tuple[int, bool], Tuple[int, int, bool]]:
    """
    Normalize an index or range of indexes for derivation.

    :param index: The index or range of indexes to normalize.
                  Can be a single non-negative integer, a string representing a single index or range,
                  or a tuple of two integers representing a range.
                  For strings, the format should be "{non-negative-number}" or "{number}-{number}".
    :type index: Union[str, int, Tuple[int, int]]
    :param hardened: Whether the index is hardened (default is False).
    :type hardened: bool

    :return: A tuple representing the normalized index or range of indexes, optionally with hardened flag.
             For a single index: (index, hardened)
             For a range of indexes: (from_index, to_index, hardened)
    :rtype: Union[Tuple[int, bool], Tuple[int, int, bool]]
    """

    if isinstance(index, tuple):
        if len(index) != 2:
            raise DerivationError(
                f"Bad index length", expected=2, got=len(index)
            )
        elif not isinstance(index[0], int) or not isinstance(index[1], int):
            raise DerivationError(
                f"Invalid index types",
                expected="both indexes must be integer instance",
                got=f"{type(index[0])}-{type(index[0])}"
            )
        elif index[0] < 0 or index[1] < 0:
            raise DerivationError(
                f"Bad index format", expected="both must be non-negative-numbers", got=index
            )
        elif index[0] > index[1]:
            raise DerivationError(
                f"Bad index, from {index[0]} index should be less than to {index[1]} index"
            )
        return *index, hardened

    elif isinstance(index, str):

        match: re.Match = re.match(
            r"^(\d+)(-(\d+))?$", index
        )
        if match:
            from_index: int = int(match.group(1))
            to_index: Optional[int] = (
                int(match.group(3)) if match.group(3) else None
            )
            if to_index is None:
                return from_index, hardened
            if from_index > to_index:
                raise DerivationError(
                    f"Bad index, from {from_index} index should be less than to {to_index} index"
                )
            return from_index, to_index, hardened
        raise DerivationError(
            f"Bad index format", expected="{non-negative-number} | {number}-{number}", got=index
        )

    elif isinstance(index, int):
        if index < 0:
            raise DerivationError(
                f"Bad index format", expected="non-negative-number", got=index
            )
        return index, hardened

    raise DerivationError(
        f"Invalid index instance", expected=(str, int, tuple), got=type(index)
    )


def normalize_derivation(
    path: Optional[str] = None, indexes: Optional[List[int]] = None
) -> Tuple[str, List[int], List[tuple]]:
    """
    Normalize a derivation path string or indexes into a consistent format.

    :param path: The derivation path string to normalize, e.g., "m/0'/1/2'/2".
                 If provided, `indexes` should be None.
                 If path is None, a default path "m/" is returned.
    :type path: Optional[str]
    :param indexes: A list of indexes to convert into a derivation path string.
                    If provided, `path` should be None.
    :type indexes: Optional[List[int]]

    :return: A tuple containing the normalized derivation path string,
             list of indexes, and list of tuples representing derivation steps.
             The tuple structure:
             (normalized_path, normalized_indexes, derivations)
             where derivations is a list of tuples (index, hardened) or (from_index, to_index, hardened).
    :rtype: Tuple[str, List[int], List[tuple]]
    """

    _path: str = "m"
    _indexes: List[int] = []
    _derivations: List[tuple] = []

    if indexes:
        path = indexes_to_path(indexes=indexes)
    elif path:
        if path in ["m", "m/"]:
            return f"{_path}/", _indexes, _derivations
        elif path[0:2] != "m/":
            raise DerivationError(
                f"Bad path format", expected="like this type of path \"m/0'/0\"", got=path
            )
    elif not path:
        return f"{_path}/", _indexes, _derivations

    for depth, index in enumerate(path.lstrip("m/").split("/")):
        if "'" in index:
            if "-" in index:
                _from_index, _to_index = index[:-1].split("-")
                _index: int = int(_to_index)
                if int(_from_index) >= int(_to_index):
                    raise DerivationError(
                        f"On {depth} depth, the starting {_from_index} must be less than the ending {_to_index} index"
                    )
                _derivations.append((int(_from_index), int(_to_index), True))
            else:
                _index: int = int(index[:-1])
                _derivations.append((_index, True))
            _indexes.append(_index + 0x80000000)
            _path += f"/{_index}'"
        else:
            if "-" in index:
                _from_index, _to_index = index.split("-")
                _index: int = int(_to_index)
                if int(_from_index) >= int(_to_index):
                    raise DerivationError(
                        f"On {depth} depth, the starting {_from_index} must be less than the ending {_to_index} index"
                    )
                _derivations.append((int(_from_index), int(_to_index), False))
            else:
                _index: int = int(index)
                _derivations.append((_index, False))
            _indexes.append(_index)
            _path += f"/{_index}"

    return _path, _indexes, _derivations


def index_tuple_to_integer(index: Union[Tuple[int, bool], Tuple[int, int, bool]]) -> int:
    """
    Convert a tuple representing an index or range of indexes into a single integer.

    :param index: The tuple representing an index or range of indexes.
                  For a single index: (index, hardened)
                  For a range of indexes: (from_index, to_index, hardened)
    :type index: Union[Tuple[int, bool], Tuple[int, int, bool]]

    :return: The integer representation of the index, with hardening flag applied if present.
    :rtype: int
    """

    if not isinstance(index, tuple):
        raise DerivationError("Invalid index instance", expected=tuple, got=type(index))
    elif len(index) == 3:
        return (index[1] + 0x80000000) if index[2] else index[0]
    elif len(index) == 2:
        return (index[0] + 0x80000000) if index[1] else index[0]
    raise DerivationError("Wrong index length", expected=[2, 3], got=len(index))


def index_tuple_to_string(index: Union[Tuple[int, bool], Tuple[int, int, bool]]) -> str:
    """
    Convert a tuple representing an index or range of indexes into a string representation.

    :param index: The tuple representing an index or range of indexes.
                  For a single index: (index, hardened)
                  For a range of indexes: (from_index, to_index, hardened)
    :type index: Union[Tuple[int, bool], Tuple[int, int, bool]]

    :return: The string representation of the index or range of indexes.
    :rtype: str
    """

    if not isinstance(index, tuple):
        raise DerivationError("Invalid index instance", expected=tuple, got=type(index))
    elif len(index) == 3:
        _from_index, _to_index, _hardened = index[0], index[1], "'" if index[2] else ""
        return f"{_from_index}-{_to_index}{_hardened}"
    elif len(index) == 2:
        _index, _hardened = index[0], "'" if index[1] else ""
        return f"{_index}{_hardened}"
    raise DerivationError("Wrong index length", expected=[2, 3], got=len(index))


def index_string_to_tuple(index: str) -> Tuple[int, bool]:
    """
    Convert a string representation of an index into a tuple.

    :param index: The string representation of the index, which may include a trailing apostrophe (').
    :type index: str

    :return: A tuple representing the index and whether it is hardened (True) or not (False).
    :rtype: Tuple[int, bool]
    """

    index_split: List[str] = index.split("'")
    return (
        (int(index_split[0]), True)
        if index.endswith("'") else
        (int(index_split[0]), False)
    )


def xor(data_1: bytes, data_2: bytes) -> bytes:
    """
    Perform bitwise XOR operation between two bytes objects.

    :param data_1: The first bytes object for XOR operation.
    :type data_1: bytes
    :param data_2: The second bytes object for XOR operation. It must be of the same length as data_1.
    :type data_2: bytes

    :return: The result of XOR operation as a bytes object.
    :rtype: bytes
    """

    return bytes(
        [b1 ^ b2 for b1, b2 in zip(data_1, data_2)]
    )


def add_no_carry(data_1: bytes, data_2: bytes) -> bytes:
    """
    Perform addition without carry between two bytes objects.

    :param data_1: The first bytes object for addition without carry.
    :type data_1: bytes
    :param data_2: The second bytes object for addition without carry. It must be of the same length as data_1.
    :type data_2: bytes

    :return: The result of addition without carry as a bytes object.
    :rtype: bytes
    """

    return bytes(
        [(b1 + b2) & 0xFF for b1, b2 in zip(data_1, data_2)]
    )


def multiply_scalar_no_carry(data: bytes, scalar: int) -> bytes:
    """
    Multiply each byte in a bytes object by a scalar without carry.

    :param data: The bytes object to multiply.
    :type data: bytes
    :param scalar: The scalar value to multiply each byte by.
    :type scalar: int

    :return: The result of multiplying each byte by the scalar, without carry, as a bytes object.
    :rtype: bytes
    """

    return bytes(
        [(b * scalar) & 0xFF for b in data]
    )


def is_bits_set(value: int, bit_num: int) -> bool:
    """
    Check if a specific bit in an integer value is set (i.e., equals 1).

    :param value: The integer value to check.
    :type value: int
    :param bit_num: The bit number to check (0-indexed, from the right).
    :type bit_num: int

    :return: True if the specified bit in `value` is set (equals 1), False otherwise.
    :rtype: bool
    """

    return (value & (1 << bit_num)) != 0


def are_bits_set(value: int, bit_mask: int) -> bool:
    """
    Check if specific bits in an integer value are set according to a bitmask.

    :param value: The integer value to check.
    :type value: int
    :param bit_mask: The bitmask representing which bits to check. Bits set to 1 in the bitmask will be checked in `value`.
    :type bit_mask: int

    :return: True if any of the bits set in `bit_mask` are also set in `value`, False otherwise.
    :rtype: bool
    """

    return (value & bit_mask) != 0


def set_bit(value: int, bit_num: int) -> int:
    """
    Set a specific bit in an integer value to 1.

    :param value: The integer value in which to set the bit.
    :type value: int
    :param bit_num: The bit number to set (0-indexed, from the right).
    :type bit_num: int

    :return: The integer value with the specified bit set to 1.
    :rtype: int
    """

    return value | (1 << bit_num)


def set_bits(value: int, bit_mask: int) -> int:
    """
    Set specific bits in an integer value according to a bitmask.

    :param value: The integer value in which to set the bits.
    :type value: int
    :param bit_mask: The bitmask representing which bits to set. Bits set to 1 in the bitmask will be set in `value`.
    :type bit_mask: int

    :return: The integer value with the specified bits set according to the bitmask.
    :rtype: int
    """

    return value | bit_mask


def reset_bit(value: int, bit_num: int) -> int:
    """
    Reset (clear) a specific bit in an integer value to 0.

    :param value: The integer value in which to reset the bit.
    :type value: int
    :param bit_num: The bit number to reset (0-indexed, from the right).
    :type bit_num: int

    :return: The integer value with the specified bit reset to 0.
    :rtype: int
    """

    return value & ~(1 << bit_num)


def reset_bits(value: int, bit_mask: int) -> int:
    """
    Reset (clear) specific bits in an integer value according to a bitmask.

    :param value: The integer value in which to reset the bits.
    :type value: int
    :param bit_mask: The bitmask representing which bits to reset. Bits set to 1 in the bitmask will be cleared in `value`.
    :type bit_mask: int

    :return: The integer value with the specified bits reset according to the bitmask.
    :rtype: int
    """

    return value & ~bit_mask


def get_bytes(data: AnyStr, unhexlify: bool = True) -> bytes:
    """
    Convert input data to bytes format.

    :param data: The input data to convert. Can be bytes or string.
    :type data: Union[bytes, str]
    :param unhexlify: Flag indicating whether to interpret strings as hexadecimal (default True).
    :type unhexlify: bool

    :return: The input data converted to bytes format.
    :rtype: bytes
    """

    if not data:
        return b''
    if isinstance(data, bytes):
        return data
    elif isinstance(data, str):
        if unhexlify:
            return bytes.fromhex(data)
        else:
            return bytes(data, 'utf-8')
    else:
        raise TypeError("Agreement must be either 'bytes' or 'string'!")


def bytes_reverse(data: bytes) -> bytes:
    """
    Reverse the order of bytes in a bytes object.

    :param data: The bytes object to reverse.
    :type data: bytes

    :return: The bytes object with its byte order reversed.
    :rtype: bytes
    """

    tmp = bytearray(data)
    tmp.reverse()
    return bytes(tmp)


def bytes_to_string(data: Union[bytes, str]) -> str:
    """
    Convert bytes or string data to a hexadecimal string representation.

    :param data: The bytes or string data to convert to hexadecimal string.
    :type data: Union[bytes, str]

    :return: The hexadecimal string representation of the input data.
    :rtype: str
    """

    if not data:
        return ''
    try:
        bytes.fromhex(data)
        return data
    except (ValueError, TypeError):
        pass
    if not isinstance(data, bytes):
        data = bytes(data, 'utf-8')
    return data.hex()


def bytes_to_integer(data: bytes, endianness: Literal["little", "big"] = "big", signed: bool = False) -> int:
    """
    Convert bytes to an integer based on specified endianness and signedness.

    :param data: The bytes object to convert to an integer.
    :type data: bytes
    :param endianness: The byte order ("little" or "big").
    :type endianness: Literal["little", "big"]
    :param signed: Flag indicating whether the integer is signed (default False).
    :type signed: bool

    :return: The integer value converted from bytes.
    :rtype: int
    """

    return int.from_bytes(data, byteorder=endianness, signed=signed)


def integer_to_bytes(data: int, bytes_num: Optional[int] = None, endianness: Literal["little", "big"] = "big", signed: bool = False) -> bytes:
    """
    Convert an integer to bytes based on specified parameters.

    :param data: The integer to convert to bytes.
    :type data: int
    :param bytes_num: Optional number of bytes to use for the conversion. If not provided, it is calculated based on the integer's bit length.
    :type bytes_num: Optional[int]
    :param endianness: The byte order ("little" or "big").
    :type endianness: Literal["little", "big"]
    :param signed: Flag indicating whether the integer is signed (default False).
    :type signed: bool

    :return: The bytes object representing the integer.
    :rtype: bytes
    """

    bytes_num = bytes_num or ((data.bit_length() if data > 0 else 1) + 7) // 8
    return data.to_bytes(bytes_num, byteorder=endianness, signed=signed)


def integer_to_binary_string(data: int, zero_pad_bit_len: int = 0) -> str:
    """
    Convert an integer to a binary string representation.

    :param data: The integer to convert to binary string.
    :type data: int
    :param zero_pad_bit_len: Optional number of bits to zero-pad the binary string (default 0).
    :type zero_pad_bit_len: int

    :return: The binary string representation of the integer.
    :rtype: str
    """

    return bin(data)[2:].zfill(zero_pad_bit_len)


def binary_string_to_integer(data: Union[bytes, str]) -> int:
    """
    Convert a binary string representation to an integer.

    :param data: The binary string or bytes object to convert to an integer.
    :type data: Union[bytes, str]

    :return: The integer value converted from the binary representation.
    :rtype: int
    """

    return int((data.encode("utf-8") if isinstance(data, str) else data), 2)


def bytes_to_binary_string(data: bytes, zero_pad_bit_len: int = 0) -> str:
    """
    Convert bytes to a binary string representation.

    :param data: The bytes object to convert to binary string.
    :type data: bytes
    :param zero_pad_bit_len: Optional number of bits to zero-pad the binary string (default 0).
    :type zero_pad_bit_len: int

    :return: The binary string representation of the bytes.
    :rtype: str
    """

    return integer_to_binary_string(bytes_to_integer(data), zero_pad_bit_len)


def binary_string_to_bytes(data: Union[bytes, str], zero_pad_byte_len: int = 0) -> bytes:
    """
    Convert a binary string representation to bytes.

    :param data: The binary string or bytes object to convert to bytes.
    :type data: Union[bytes, str]
    :param zero_pad_byte_len: Optional number of bytes to zero-pad the resulting bytes object (default 0).
    :type zero_pad_byte_len: int

    :return: The bytes object converted from the binary representation.
    :rtype: bytes
    """

    return binascii.unhexlify(hex(binary_string_to_integer(data))[2:].zfill(zero_pad_byte_len))


def decode(data: Union[bytes, str], encoding: str = "utf-8") -> str:
    """
    Decode bytes or return a string unchanged.

    :param data: The bytes or string data to decode.
    :type data: Union[bytes, str]
    :param encoding: The encoding to use when decoding bytes (default is 'utf-8').
    :type encoding: str

    :return: The decoded string if data is bytes; otherwise, returns the input string unchanged.
    :rtype: str
    """

    if isinstance(data, str):
        return data
    if isinstance(data, bytes):
        return data.decode(encoding)
    raise TypeError("Invalid data type")


def encode(data: Union[bytes, str], encoding: str = "utf-8") -> bytes:
    """
    Encode string or return bytes unchanged.

    :param data: The string or bytes data to encode.
    :type data: Union[bytes, str]
    :param encoding: The encoding to use when encoding a string to bytes (default is 'utf-8').
    :type encoding: str

    :return: The encoded bytes if data is string; otherwise, returns the input bytes unchanged.
    :rtype: bytes
    """

    if isinstance(data, str):
        return data.encode(encoding)
    if isinstance(data, bytes):
        return data
    raise TypeError("Invalid data type")


def convert_bits(
    data: Union[bytes, List[int]], from_bits: int, to_bits: int
) -> Optional[List[int]]:
    """
    Convert data represented by 'from_bits' into 'to_bits' per element.

    :param data: The data to convert, represented as bytes or a list of integers.
    :type data: Union[bytes, List[int]]
    :param from_bits: Number of bits each element of 'data' represents initially.
    :type from_bits: int
    :param to_bits: Number of bits each element of the result should represent after conversion.
    :type to_bits: int

    :return: The converted data as a list of integers, or None if conversion fails.
    :rtype: Optional[List[int]]
    """

    max_out_val = (1 << to_bits) - 1

    acc = 0
    bits = 0
    ret = []

    for value in data:
        if value < 0 or (value >> from_bits):
            return None
        acc |= value << bits
        bits += from_bits
        while bits >= to_bits:
            ret.append(acc & max_out_val)
            acc = acc >> to_bits
            bits -= to_bits

    if bits != 0:
        ret.append(acc & max_out_val)

    return ret


def bytes_chunk_to_words(
    bytes_chunk: bytes, words_list: List[str], endianness: Literal["little", "big"]
) -> List[str]:
    """
    Convert a bytes chunk into a list of words based on a given word list and endianness.

    :param bytes_chunk: The bytes chunk to convert into words.
    :type bytes_chunk: bytes
    :param words_list: The list of words to choose from when converting.
    :type words_list: List[str]
    :param endianness: The endianness to use when interpreting the bytes chunk ("little" or "big").
    :type endianness: Literal["little", "big"]

    :return: A list of three words selected from words_list based on the bytes chunk.
    :rtype: List[str]
    """

    words_list_length = len(words_list)

    chunk: int = bytes_to_integer(bytes_chunk, endianness=endianness)

    word_1_index = chunk % words_list_length
    word_2_index = ((chunk // words_list_length) + word_1_index) % words_list_length
    word_3_index = ((chunk // words_list_length // words_list_length) + word_2_index) % words_list_length

    return [words_list[index] for index in (word_1_index, word_2_index, word_3_index)]


def words_to_bytes_chunk(
    word_1: str, word_2: str, word_3: str, words_list: List[str], endianness: Literal["little", "big"]
) -> bytes:
    """
    Convert three words into a bytes chunk based on a given word list and endianness.

    :param word_1: The first word to convert.
    :type word_1: str
    :param word_2: The second word to convert.
    :type word_2: str
    :param word_3: The third word to convert.
    :type word_3: str
    :param words_list: The list of words from which to select the words.
    :type words_list: List[str]
    :param endianness: The endianness to use when encoding the chunk into bytes ("little" or "big").
    :type endianness: Literal["little", "big"]

    :return: The bytes chunk representing the three words.
    :rtype: bytes
    """

    words_list_length = len(words_list)
    words_list_with_index: dict = {
        words_list[i]: i for i in range(len(words_list))
    }

    word_1_index, word_2_index,  word_3_index = (
        words_list_with_index[word_1], words_list_with_index[word_2] % words_list_length, words_list_with_index[word_3] % words_list_length
    )

    chunk: int = (
        word_1_index + (
            words_list_length * ((word_2_index - word_1_index) % words_list_length)
        ) + (
            words_list_length * words_list_length * ((word_3_index - word_2_index) % words_list_length)
        )
    )

    return integer_to_bytes(
        chunk, bytes_num=4, endianness=endianness
    )



#!/usr/bin/env python3

# Copyright © 2020-2024, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import (
    Union, Dict, List
)

import unicodedata

from ...entropies import (
    IEntropy, MoneroEntropy, MONERO_ENTROPY_STRENGTHS
)
from ...crypto import crc32
from ...exceptions import (
    Error, EntropyError, MnemonicError, ChecksumError
)
from ...utils import (
    get_bytes, bytes_to_string, bytes_to_integer, bytes_chunk_to_words, words_to_bytes_chunk
)
from ..imnemonic import IMnemonic


class MONERO_MNEMONIC_WORDS:
    """
    Constants for Monero mnemonic words.
    """

    TWELVE: int = 12
    THIRTEEN: int = 13
    TWENTY_FOUR: int = 24
    TWENTY_FIVE: int = 25


class MONERO_MNEMONIC_LANGUAGES:
    """
    Constants for Monero mnemonic languages.
    """

    CHINESE_SIMPLIFIED: str = "chinese-simplified"
    DUTCH: str = "dutch"
    ENGLISH: str = "english"
    FRENCH: str = "french"
    GERMAN: str = "german"
    ITALIAN: str = "italian"
    JAPANESE: str = "japanese"
    PORTUGUESE: str = "portuguese"
    RUSSIAN: str = "russian"
    SPANISH: str = "spanish"


class MoneroMnemonic(IMnemonic):
    """
    Designed for Monero's mnemonic system, focusing on privacy and security,
    enabling users to recover their wallets through unique seed phrases.

    Here are available ``MONERO_MNEMONIC_WORDS``:

    +-----------------------+----------------------+
    | Name                  | Value                |
    +=======================+======================+
    | TWELVE                | 12                   |
    +-----------------------+----------------------+
    | THIRTEEN              | 13                   |
    +-----------------------+----------------------+
    | TWENTY_FOUR           | 24                   |
    +-----------------------+----------------------+
    | TWENTY_FIVE           | 25                   |
    +-----------------------+----------------------+

    Here are available ``MONERO_MNEMONIC_LANGUAGES``:

    +-----------------------+----------------------+
    | Name                  | Value                |
    +=======================+======================+
    | CHINESE_SIMPLIFIED    | chinese-simplified   |
    +-----------------------+----------------------+
    | DUTCH                 | dutch                |
    +-----------------------+----------------------+
    | ENGLISH               | english              |
    +-----------------------+----------------------+
    | FRENCH                | french               |
    +-----------------------+----------------------+
    | GERMAN                | chinese-german       |
    +-----------------------+----------------------+
    | ITALIAN               | italian              |
    +-----------------------+----------------------+
    | JAPANESE              | japanese             |
    +-----------------------+----------------------+
    | PORTUGUESE            | portuguese           |
    +-----------------------+----------------------+
    | RUSSIAN               | russian              |
    +-----------------------+----------------------+
    | SPANISH               | spanish              |
    +-----------------------+----------------------+
    """

    word_bit_length: int = 11
    words_list_number: int = 1626
    words_list: List[int] = [
        MONERO_MNEMONIC_WORDS.TWELVE,
        MONERO_MNEMONIC_WORDS.THIRTEEN,
        MONERO_MNEMONIC_WORDS.TWENTY_FOUR,
        MONERO_MNEMONIC_WORDS.TWENTY_FIVE
    ]
    words_checksum: List[int] = [
        MONERO_MNEMONIC_WORDS.THIRTEEN,
        MONERO_MNEMONIC_WORDS.TWENTY_FIVE
    ]
    words_to_entropy_strength: Dict[int, int] = {
        MONERO_MNEMONIC_WORDS.TWELVE: MONERO_ENTROPY_STRENGTHS.ONE_HUNDRED_TWENTY_EIGHT,
        MONERO_MNEMONIC_WORDS.THIRTEEN: MONERO_ENTROPY_STRENGTHS.ONE_HUNDRED_TWENTY_EIGHT,
        MONERO_MNEMONIC_WORDS.TWENTY_FOUR: MONERO_ENTROPY_STRENGTHS.TWO_HUNDRED_FIFTY_SIX,
        MONERO_MNEMONIC_WORDS.TWENTY_FIVE: MONERO_ENTROPY_STRENGTHS.TWO_HUNDRED_FIFTY_SIX
    }
    languages: List[str] = [
        MONERO_MNEMONIC_LANGUAGES.CHINESE_SIMPLIFIED,
        MONERO_MNEMONIC_LANGUAGES.DUTCH,
        MONERO_MNEMONIC_LANGUAGES.ENGLISH,
        MONERO_MNEMONIC_LANGUAGES.FRENCH,
        MONERO_MNEMONIC_LANGUAGES.GERMAN,
        MONERO_MNEMONIC_LANGUAGES.ITALIAN,
        MONERO_MNEMONIC_LANGUAGES.JAPANESE,
        MONERO_MNEMONIC_LANGUAGES.PORTUGUESE,
        MONERO_MNEMONIC_LANGUAGES.RUSSIAN,
        MONERO_MNEMONIC_LANGUAGES.SPANISH
    ]
    language_unique_prefix_lengths: Dict[str, int] = {
        MONERO_MNEMONIC_LANGUAGES.CHINESE_SIMPLIFIED: 1,
        MONERO_MNEMONIC_LANGUAGES.DUTCH: 4,
        MONERO_MNEMONIC_LANGUAGES.ENGLISH: 3,
        MONERO_MNEMONIC_LANGUAGES.FRENCH: 4,
        MONERO_MNEMONIC_LANGUAGES.GERMAN: 4,
        MONERO_MNEMONIC_LANGUAGES.ITALIAN: 4,
        MONERO_MNEMONIC_LANGUAGES.JAPANESE: 4,
        MONERO_MNEMONIC_LANGUAGES.PORTUGUESE: 4,
        MONERO_MNEMONIC_LANGUAGES.SPANISH: 4,
        MONERO_MNEMONIC_LANGUAGES.RUSSIAN: 4
    }
    wordlist_path: Dict[str, str] = {
        MONERO_MNEMONIC_LANGUAGES.CHINESE_SIMPLIFIED: "monero/wordlist/chinese_simplified.txt",
        MONERO_MNEMONIC_LANGUAGES.DUTCH: "monero/wordlist/dutch.txt",
        MONERO_MNEMONIC_LANGUAGES.ENGLISH: "monero/wordlist/english.txt",
        MONERO_MNEMONIC_LANGUAGES.FRENCH: "monero/wordlist/french.txt",
        MONERO_MNEMONIC_LANGUAGES.GERMAN: "monero/wordlist/german.txt",
        MONERO_MNEMONIC_LANGUAGES.ITALIAN: "monero/wordlist/italian.txt",
        MONERO_MNEMONIC_LANGUAGES.JAPANESE: "monero/wordlist/japanese.txt",
        MONERO_MNEMONIC_LANGUAGES.PORTUGUESE: "monero/wordlist/portuguese.txt",
        MONERO_MNEMONIC_LANGUAGES.RUSSIAN: "monero/wordlist/russian.txt",
        MONERO_MNEMONIC_LANGUAGES.SPANISH: "monero/wordlist/spanish.txt"
    }

    @classmethod
    def name(cls) -> str:
        """
        Get the name of the mnemonic class.

        :return: The name of the entropy class.
        :rtype: str
        """
        return "Monero"

    @classmethod
    def from_words(cls, words: int, language: str) -> str:
        """
        Generates a mnemonic phrase from a specified number of words and language.

        :param words: The number of words in the mnemonic phrase.
        :type words: int
        :param language: The language for which to generate the mnemonic phrase.
        :type language: str

        :return: The generated mnemonic phrase.
        :rtype: str
        """

        if words not in cls.words_list:
            raise MnemonicError("Invalid mnemonic words number", expected=cls.words_list, got=words)

        return cls.from_entropy(
            entropy=MoneroEntropy.generate(cls.words_to_entropy_strength[words]),
            language=language,
            checksum=(
                True if words in cls.words_checksum else False
            )
        )

    @classmethod
    def from_entropy(cls, entropy: Union[str, bytes, IEntropy], language: str, checksum: bool = False) -> str:
        """
        Generates a mnemonic phrase from entropy data.

        :param entropy: The entropy data used to generate the mnemonic phrase.
        :type entropy: Union[str, bytes, IEntropy]
        :param language: The language for which to generate the mnemonic phrase.
        :type language: str
        :param checksum: Whether to include a checksum in the mnemonic phrase.
        :type checksum: bool

        :return: The generated mnemonic phrase.
        :rtype: str
        """
        if isinstance(entropy, str) or isinstance(entropy, bytes):
            return cls.encode(
                entropy=entropy, language=language, checksum=checksum
            )
        elif isinstance(entropy, MoneroEntropy):
            return cls.encode(
                entropy=entropy.entropy(), language=language, checksum=checksum
            )
        raise EntropyError(
            "Invalid entropy instance", expected=[str, bytes, MoneroEntropy], got=type(entropy)
        )

    @classmethod
    def encode(cls, entropy: Union[str, bytes], language: str, checksum: bool = False) -> str:
        """
        Generates a mnemonic phrase from entropy data.

        :param entropy: The entropy data used to generate the mnemonic phrase.
        :type entropy: Union[str, bytes]
        :param language: The language for which to generate the mnemonic phrase.
        :type language: str
        :param checksum: Whether to include a checksum in the mnemonic phrase.
        :type checksum: bool

        :return: The generated mnemonic phrase.
        :rtype: str
        """

        entropy: bytes = get_bytes(entropy)
        if not MoneroEntropy.is_valid_bytes_strength(len(entropy)):
            raise EntropyError(
                "Wrong entropy strength", expected=MoneroEntropy.strengths, got=(len(entropy) * 8)
            )

        mnemonic: List[str] = []
        words_list: List[str] = cls.normalize(cls.get_words_list_by_language(language=language))
        if len(words_list) != cls.words_list_number:
            raise Error(
                "Invalid number of loaded words list", expected=cls.words_list_number, got=len(words_list)
            )

        for index in range(len(get_bytes(entropy)) // 4):
            mnemonic += bytes_chunk_to_words(
                entropy[index * 4:(index * 4) + 4], words_list, "little"
            )

        if checksum:
            unique_prefix_length = cls.language_unique_prefix_lengths[language]
            prefixes = "".join(word[:unique_prefix_length] for word in mnemonic)
            checksum_word = mnemonic[
                bytes_to_integer(crc32(prefixes)) % len(mnemonic)
            ]
            mnemonic = mnemonic + [checksum_word]

        return " ".join(cls.normalize(mnemonic))

    @classmethod
    def decode(cls, mnemonic: str, **kwargs) -> str:
        """
        Decodes a mnemonic phrase into entropy data.

        :param mnemonic: The mnemonic phrase to decode.
        :type mnemonic: str
        :param kwargs: Additional keyword arguments (language, checksum).

        :return: The decoded entropy data.
        :rtype: str
        """

        words: list = cls.normalize(mnemonic)
        if len(words) not in cls.words_list:
            raise MnemonicError("Invalid mnemonic words count", expected=cls.words_list, got=len(words))

        words_list, language = cls.find_language(mnemonic=words)
        if len(words_list) != cls.words_list_number:
            raise Error(
                "Invalid number of loaded words list", expected=cls.words_list_number, got=len(words_list)
            )

        if len(words) in cls.words_checksum:
            mnemonic: list = words[:-1]
            unique_prefix_length = cls.language_unique_prefix_lengths[language]
            prefixes = "".join(word[:unique_prefix_length] for word in mnemonic)
            checksum_word = mnemonic[
                bytes_to_integer(crc32(prefixes)) % len(mnemonic)
            ]
            if words[-1] != checksum_word:
                raise ChecksumError(
                    "Invalid checksum", expected=checksum_word, got=words[-1]
                )

        entropy: bytes = b""
        for index in range(len(words) // 3):
            word_1, word_2, word_3 = words[index * 3:(index * 3) + 3]
            entropy += words_to_bytes_chunk(
                word_1, word_2, word_3, words_list, "little"
            )
        return bytes_to_string(entropy)

    @classmethod
    def normalize(cls, mnemonic: Union[str, List[str]]) -> List[str]:
        """
        Normalizes the given mnemonic by splitting it into a list of words if it is a string.

        :param mnemonic: The mnemonic value, which can be a single string of words or a list of words.
        :type mnemonic: Union[str, List[str]]

        :return: A list of words from the mnemonic.
        :rtype: List[str]
        """

        mnemonic: list = mnemonic.split() if isinstance(mnemonic, str) else mnemonic
        return list(map(lambda _: unicodedata.normalize("NFKD", _.lower()), mnemonic))



#!/usr/bin/env python3

# Copyright © 2020-2024, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from .ientropy import IEntropy


class ALGORAND_ENTROPY_STRENGTHS:
    """
    Constants representing the entropy strengths for Algorand.
    """

    TWO_HUNDRED_FIFTY_SIX: int = 256


class AlgorandEntropy(IEntropy):
    """
    Uses entropy to generate a mnemonic phrase specific to Algorand,
    ensuring secure account creation with a unique checksum for
    address verification.

    .. note::
        This class inherits from the ``IEntropy`` class, thereby ensuring that all functions are accessible.

    Here are available ``ALGORAND_ENTROPY_STRENGTHS``:

    +-----------------------+-------+
    | Name                  | Value |
    +=======================+=======+
    | TWO_HUNDRED_FIFTY_SIX | 256   |
    +-----------------------+-------+
    """

    strengths = [
        ALGORAND_ENTROPY_STRENGTHS.TWO_HUNDRED_FIFTY_SIX
    ]

    @classmethod
    def name(cls) -> str:
        """
        Get the name of the entropy class.

        :return: The name of the entropy class.
        :rtype: str
        """

        return "Algorand"



#!/usr/bin/env python3

# Copyright © 2020-2024, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

import json
import click
import sys

from ...mnemonics import MNEMONICS
from ...seeds import (
    ISeed, BIP39Seed, CardanoSeed, ElectrumV2Seed, SEEDS
)


def generate_seed(**kwargs) -> None:
    try:
        if not kwargs.get("mnemonic"):
            click.echo(click.style(f"Mnemonic is required for {kwargs.get('client')} client"), err=True)
            sys.exit()

        if not SEEDS.is_seed(name=kwargs.get("client")):
            click.echo(click.style(
                f"Wrong seed client, (expected={SEEDS.names()}, got='{kwargs.get('client')}')"
            ), err=True)
            sys.exit()

        if kwargs.get("client") == "Electrum-V2":
            if not MNEMONICS.mnemonic(name="Electrum-V2").is_valid(
                mnemonic=kwargs.get("mnemonic"), mnemonic_type=kwargs.get("mnemonic_type")
            ):
                click.echo(click.style(f"Invalid Electrum-V2 mnemonic"), err=True)
                sys.exit()
        else:
            mnemonic_name: str = "BIP39" if kwargs.get("client") == CardanoSeed.name() else kwargs.get("client")
            if not MNEMONICS.mnemonic(name=mnemonic_name).is_valid(mnemonic=kwargs.get("mnemonic")):
                click.echo(click.style(f"Invalid {mnemonic_name} mnemonic"), err=True)
                sys.exit()


        if kwargs.get("client") == BIP39Seed.name():
            seed: ISeed = BIP39Seed(
                seed=BIP39Seed.from_mnemonic(
                    mnemonic=kwargs.get("mnemonic"),
                    passphrase=kwargs.get("passphrase")
                )
            )
        elif kwargs.get("client") == CardanoSeed.name():
            seed: ISeed = CardanoSeed(
                seed=CardanoSeed.from_mnemonic(
                    mnemonic=kwargs.get("mnemonic"),
                    passphrase=kwargs.get("passphrase"),
                    cardano_type=kwargs.get("cardano_type")
                )
            )
        elif kwargs.get("client") == ElectrumV2Seed.name():
            seed: ISeed = ElectrumV2Seed(
                seed=ElectrumV2Seed.from_mnemonic(
                    mnemonic=kwargs.get("mnemonic"),
                    passphrase=kwargs.get("passphrase"),
                    mnemonic_type=kwargs.get("mnemonic_type")
                )
            )
        else:
            seed: ISeed = SEEDS.seed(name=kwargs.get("client")).__call__(
                seed=SEEDS.seed(name=kwargs.get("client")).from_mnemonic(
                    mnemonic=kwargs.get("mnemonic")
                )
            )
        output: dict = {
            "client": seed.name(),
            "seed": seed.seed()
        }
        if seed.name() == CardanoSeed.name():
            output["cardano_type"] = kwargs.get("cardano_type")
        elif seed.name() == ElectrumV2Seed.name():
            output["mnemonic_type"] = kwargs.get("mnemonic_type")
        click.echo(json.dumps(
            output, indent=kwargs.get("indent", 4), ensure_ascii=kwargs.get("ensure_ascii", False)
        ))

    except Exception as exception:
        click.echo(click.style(f"Error: {str(exception)}"), err=True)
        sys.exit()



#!/usr/bin/env python3

# Copyright © 2020-2024, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import (
    Tuple, Union, Optional, Dict
)

from ..utils import (
    normalize_index, normalize_derivation, index_tuple_to_string
)
from ..exceptions import DerivationError
from ..cryptocurrencies import Cardano
from .iderivation import IDerivation


class ROLES:

    EXTERNAL_CHAIN: str = "external-chain"
    INTERNAL_CHAIN: str = "internal-chain"
    STAKING_KEY: str = "staking-key"


class CIP1852Derivation(IDerivation):  # https://github.com/cardano-foundation/CIPs/blob/master/CIP-1852/README.md
    """
    This class implements the CIP1852 standard for hierarchical deterministic wallets.
    CIP1852 defines a specific path structure for deriving keys from a master seed.

    .. note::
        This class inherits from the ``IDerivation`` class, thereby ensuring that all functions are accessible.

    +-----------------------+------------------+
    | Name                  | Value            |
    +=======================+==================+
    | EXTERNAL_CHAIN        | external-chain   |
    +-----------------------+------------------+
    | INTERNAL_CHAIN        | internal-chain   |
    +-----------------------+------------------+
    | STAKING_KEY           | staking-key      |
    +-----------------------+------------------+
    """

    _purpose: Tuple[int, bool] = (1852, True)

    _coin_type: Tuple[int, bool]
    _account: Union[Tuple[int, bool], Tuple[int, int, bool]]
    _role: Tuple[int, bool]
    _address: Union[Tuple[int, bool], Tuple[int, int, bool]]

    def __init__(
        self,
        coin_type: Union[str, int] = Cardano.COIN_TYPE,
        account: Union[str, int, Tuple[int, int]] = 0,
        role: Union[str, int] = ROLES.EXTERNAL_CHAIN,
        address: Union[str, int, Tuple[int, int]] = 0
    ) -> None:
        """
        Initializes a CIP1852Derivation object with specified parameters.

        :param coin_type: The index for the coin type derivation. Default is 1815.
        :type coin_type: Union[str, int]
        :param account: The account index for the derivation path. Default is 0.
        :type account: Union[str, int, Tuple[int, int]]
        :param role: The role index for the derivation path. Default is "external-chain".
                     Must be one of the keys in `self.roles` or 0, 1, 2.
        :type role: Union[str, int]
        :param address: The address index for the derivation path. Default is 0.
        :type address: Union[str, int, Tuple[int, int]]
        """
        super(CIP1852Derivation, self).__init__()

        self._coin_type = normalize_index(index=coin_type, hardened=True)
        self._account = normalize_index(index=account, hardened=True)
        self._role = normalize_index(
            index=self.get_role_value(role=role, name_only=False), hardened=False
        )
        self._address = normalize_index(index=address, hardened=False)
        self._path, self._indexes, self._derivations = normalize_derivation(path=(
            f"m/{index_tuple_to_string(index=self._purpose)}/"
            f"{index_tuple_to_string(index=self._coin_type)}/"
            f"{index_tuple_to_string(index=self._account)}/"
            f"{index_tuple_to_string(index=self._role)}/"
            f"{index_tuple_to_string(index=self._address)}"
        ))
        
    @classmethod
    def name(cls) -> str:
        """
        Get the name of the derivation class.

        :return: The name of the derivation class.
        :rtype: str
        """

        return "CIP1852"

    def get_role_value(self, role: Union[str, int], name_only: bool = False):
        if isinstance(role, (list, tuple)):
            raise DerivationError(
                "Bad role instance", expected="int | str", got=type(role).__name__
            )
        external_change = [ROLES.EXTERNAL_CHAIN, 0, '0']
        internal_change = [ROLES.INTERNAL_CHAIN, 1, '1']
        staking_key = [ROLES.STAKING_KEY, 2, '2']
        expected_role = external_change + internal_change + staking_key
        if role not in expected_role:
            raise DerivationError(
                f"Bad {self.name()} role index", expected=expected_role, got=role
            )
        if role in external_change:
            return ROLES.EXTERNAL_CHAIN if name_only else 0
        if role in internal_change:
            return ROLES.INTERNAL_CHAIN if name_only else 1
        if role in staking_key:
            return ROLES.STAKING_KEY if name_only else 2

    def from_coin_type(self, coin_type: Union[str, int]) -> "CIP1852Derivation":
        """
        Sets the derivation path based on the provided coin type index.

        :param coin_type: The index for the coin type derivation.
        :type coin_type: Union[str, int]

        :return: The updated instance with the new derivation path based on the coin type index.
        :rtype: CIP1852Derivation
        """
        self._coin_type = normalize_index(index=coin_type, hardened=True)
        self._path, self._indexes, self._derivations = normalize_derivation(path=(
            f"m/{index_tuple_to_string(index=self._purpose)}/"
            f"{index_tuple_to_string(index=self._coin_type)}/"
            f"{index_tuple_to_string(index=self._account)}/"
            f"{index_tuple_to_string(index=self._role)}/"
            f"{index_tuple_to_string(index=self._address)}"
        ))
        return self

    def from_account(self, account: Union[str, int, Tuple[int, int]]) -> "CIP1852Derivation":
        """
        Sets the derivation path based on the provided account index.

        :param account: The index for the account derivation.
        :type account: Union[str, int, Tuple[int, int]]

        :return: The updated instance with the new derivation path based on the account index.
        :rtype: CIP1852Derivation
        """

        self._account = normalize_index(index=account, hardened=True)
        self._path, self._indexes, self._derivations = normalize_derivation(path=(
            f"m/{index_tuple_to_string(index=self._purpose)}/"
            f"{index_tuple_to_string(index=self._coin_type)}/"
            f"{index_tuple_to_string(index=self._account)}/"
            f"{index_tuple_to_string(index=self._role)}/"
            f"{index_tuple_to_string(index=self._address)}"
        ))
        return self

    def from_role(self, role: Union[str, int]) -> "CIP1852Derivation":
        """
        Sets the derivation path based on the provided role index.

        :param role: The index for the role derivation.
        :type role: Union[str, int]

        :return: The updated instance with the new derivation path based on the role index.
        :rtype: CIP1852Derivation
        """

        self._role = normalize_index(
            index=self.get_role_value(role=role, name_only=False), hardened=False
        )
        self._path, self._indexes, self._derivations = normalize_derivation(path=(
            f"m/{index_tuple_to_string(index=self._purpose)}/"
            f"{index_tuple_to_string(index=self._coin_type)}/"
            f"{index_tuple_to_string(index=self._account)}/"
            f"{index_tuple_to_string(index=self._role)}/"
            f"{index_tuple_to_string(index=self._address)}"
        ))
        return self

    def from_address(self, address: Union[str, int, Tuple[int, int]]) -> "CIP1852Derivation":
        """
        Sets the derivation path based on the provided address index.

        :param address: The index or tuple of indices for the address derivation.
        :type address: Union[str, int, Tuple[int, int]]

        :return: The updated instance with the new derivation path based on the address index.
        :rtype: CIP1852Derivation
        """

        self._address = normalize_index(index=address, hardened=False)
        self._path, self._indexes, self._derivations = normalize_derivation(path=(
            f"m/{index_tuple_to_string(index=self._purpose)}/"
            f"{index_tuple_to_string(index=self._coin_type)}/"
            f"{index_tuple_to_string(index=self._account)}/"
            f"{index_tuple_to_string(index=self._role)}/"
            f"{index_tuple_to_string(index=self._address)}"
        ))
        return self

    def clean(self) -> "CIP1852Derivation":
        """
        Resets the derivation state to default values for a clean derivation path.

        :return: The updated instance with a clean derivation path.
        :rtype: CIP1852Derivation
        """

        self._account = normalize_index(index=0, hardened=True)
        self._role = normalize_index(
            index=self.get_role_value(role=ROLES.EXTERNAL_CHAIN, name_only=False), hardened=False
        )
        self._address = normalize_index(index=0, hardened=False)
        self._path, self._indexes, self._derivations = normalize_derivation(path=(
            f"m/{index_tuple_to_string(index=self._purpose)}/"
            f"{index_tuple_to_string(index=self._coin_type)}/"
            f"{index_tuple_to_string(index=self._account)}/"
            f"{index_tuple_to_string(index=self._role)}/"
            f"{index_tuple_to_string(index=self._address)}"
        ))
        return self

    def purpose(self) -> int:
        """
        Retrieves the purpose value associated with the current instance.

        :return: The purpose value as an integer.
        :rtype: int
        """

        return self._purpose[0]

    def coin_type(self) -> int:
        """
        Retrieves the coin type associated with the current instance.

        :return: The coin type integer.
        :rtype: int
        """

        return self._coin_type[0]

    def account(self) -> int:
        """
        Retrieves the account associated with the current instance.

        :return: The account integer.
        :rtype: int
        """

        return (
            self._account[1] if len(self._account) == 3 else self._account[0]
        )

    def role(self, name_only: bool = True) -> str:
        """
        Retrieves the role associated with the current instance.

        :param name_only: Return the role name if True, or its index if False.
        :type name_only: bool
        :return: The role string.
        :rtype: str
        """
        return self.get_role_value(role=self._role[0], name_only=name_only)

    def address(self) -> int:
        """
        Retrieves the address index from the internal `_address` attribute.

        :return: The address index.
        :rtype: int
        """

        return (
            self._address[1] if len(self._address) == 3 else self._address[0]
        )



#!/usr/bin/env python3

# Copyright © 2020-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import (
    Any, Union
)

from ..libs.bech32 import (
    bech32_encode, bech32_decode
)
from ..eccs import (
    IPublicKey, SLIP10Secp256k1PublicKey, validate_and_get_public_key
)
from ..cryptocurrencies import Injective
from ..utils import (
    get_bytes, bytes_to_string
)
from .ethereum import EthereumAddress
from .iaddress import IAddress


class InjectiveAddress(IAddress):

    hrp: str = Injective.NETWORKS.MAINNET.HRP

    @staticmethod
    def name() -> str:
        """
        Returns the name of the cryptocurrency, which is "Injective".

        :return: The name of the cryptocurrency.
        :rtype: str
        """

        return "Injective"

    @classmethod
    def encode(cls, public_key: Union[bytes, str, IPublicKey], **kwargs: Any) -> str:
        """
        Encodes a public key into an Injective address using Bech32 encoding.

        :param public_key: The public key to encode.
        :type public_key: Union[bytes, str, IPublicKey]
        :param kwargs: Additional keyword arguments.
        :type kwargs: Any

        :return: The encoded Injective address.
        :rtype: str
        """

        public_key: IPublicKey = validate_and_get_public_key(
            public_key=public_key, public_key_cls=SLIP10Secp256k1PublicKey
        )

        return bech32_encode(
            kwargs.get("hrp", cls.hrp), get_bytes(EthereumAddress.encode(public_key)[2:])
        )

    @classmethod
    def decode(cls, address: str, **kwargs: Any) -> str:
        """
        Decodes an Injective address from Bech32 encoding.

        :param address: The Injective address to decode.
        :type address: str
        :param kwargs: Additional keyword arguments.
        :type kwargs: Any

        :return: The decoded public key.
        :rtype: str
        """

        hrp, address_decode = bech32_decode(
            kwargs.get("hrp", cls.hrp), address
        )

        expected_length: int = 20
        if len(address_decode) != expected_length:
            raise ValueError(f"Invalid length (expected: {expected_length}, got: {len(address_decode)})")

        return bytes_to_string(address_decode)



#!/usr/bin/env python3

# Copyright © 2020-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from ....iecc import IEllipticCurveCryptography
from .. import SLIP10Ed25519ECC
from .point import SLIP10Ed25519Blake2bPoint
from .public_key import SLIP10Ed25519Blake2bPublicKey
from .private_key import SLIP10Ed25519Blake2bPrivateKey


class SLIP10Ed25519Blake2bECC(IEllipticCurveCryptography):

    NAME = "SLIP10-Ed25519-Blake2b"
    ORDER = SLIP10Ed25519ECC.ORDER
    GENERATOR = SLIP10Ed25519ECC.GENERATOR
    POINT = SLIP10Ed25519Blake2bPoint
    PUBLIC_KEY = SLIP10Ed25519Blake2bPublicKey
    PRIVATE_KEY = SLIP10Ed25519Blake2bPrivateKey



#!/usr/bin/env python3

from hashlib import sha256
from Crypto.Hash import keccak
from typing import List

import six


__base58_alphabet = "123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz"


def checksum_encode(address, crypto="eth"):
    out = ""
    keccak_256 = keccak.new(digest_bits=256)
    addr = address.lower().replace("0x", "") if crypto == "eth" else address.lower().replace("xdc", "")
    keccak_256.update(addr.encode("ascii"))
    hash_addr = keccak_256.hexdigest()
    for i, c in enumerate(addr):
        if int(hash_addr[i], 16) >= 8:
            out += c.upper()
        else:
            out += c
    return ("0x" + out) if crypto == "eth" else ("xdc" + out)


def string_to_int(data):
    val = 0

    if type(data) == str:
        data = bytearray(data)

    for (i, c) in enumerate(data[::-1]):
        val += (256 ** i) * c
    return val


def ensure_string(data):
    if isinstance(data, six.binary_type):
        return data.decode("utf-8")
    elif not isinstance(data, six.string_types):
        raise ValueError("Invalid value for string")
    return data


def encode(data, alphabet=__base58_alphabet):
    enc = ""
    val = string_to_int(data)
    while val >= len(alphabet):
        val, mod = divmod(val, len(alphabet))
        enc = alphabet[mod] + enc
    if val:
        enc = alphabet[val] + enc

    n = len(data) - len(data.lstrip(b"\0"))
    return alphabet[0] * n + enc


def check_encode(raw, alphabet=__base58_alphabet):
    chk = sha256(sha256(raw).digest()).digest()[:4]
    return encode(raw + chk, alphabet)


def decode(data, alphabet=__base58_alphabet):
    if bytes != str:
        data = bytes(data, "ascii")

    val = 0
    prefix = 0
    for c in data:
        val = (val * len(alphabet)) + alphabet.encode("utf-8").find(c)
        if val == 0:
            prefix += 1

    dec = bytearray()
    while val > 0:
        val, mod = divmod(val, 256)
        dec.append(mod)

    dec.extend(bytearray(prefix))

    return bytes(dec[::-1])


def check_decode(enc, alphabet=__base58_alphabet):
    dec = decode(enc, alphabet)
    raw, chk = dec[:-4], dec[-4:]
    if chk != sha256(sha256(raw).digest()).digest()[:4]:
        raise ValueError("base58 decoding checksum error")
    else:
        return raw


def pad(enc: str, pad_len: int) -> str:
    return enc.rjust(pad_len, __base58_alphabet[0])


def encode_monero(data: bytes) -> str:
    enc = ""

    # Get lengths
    data_len: int = len(data)
    block_enc_len: int = 11
    block_dec_len: int = 8

    block_enc_bytes_lens: List[int] = [
        0, 2, 3, 5, 6, 7, 9, 10, 11
    ]

    # Compute total block count and last block length
    tot_block_cnt, last_block_enc_len = divmod(data_len, block_dec_len)

    # Encode each single block and pad
    for i in range(tot_block_cnt):
        block_enc = encode(data[i * block_dec_len:(i + 1) * block_dec_len])
        enc += pad(block_enc, block_enc_len)

    # Encode last block and pad
    if last_block_enc_len > 0:
        block_enc = encode(
            data[tot_block_cnt * block_dec_len:(tot_block_cnt * block_dec_len) + last_block_enc_len])
        enc += pad(block_enc, block_enc_bytes_lens[last_block_enc_len])

    return enc


def unpad(dec: bytes, unpad_len: int) -> bytes:
    return dec[len(dec) - unpad_len:len(dec)]


def decode_monero(data: str) -> bytes:
    dec = b""

    # Get lengths
    data_len: int = len(data)
    block_enc_len: int = 11
    block_dec_len: int = 8

    block_enc_bytes_lens: List[int] = [
        0, 2, 3, 5, 6, 7, 9, 10, 11
    ]

    # Compute block count and last block length
    tot_block_cnt, last_block_enc_len = divmod(data_len, block_enc_len)

    # Get last block decoded length
    last_block_dec_len = block_enc_bytes_lens.index(last_block_enc_len)

    # Decode each single block and unpad
    for i in range(tot_block_cnt):
        block_dec = decode(data[(i * block_enc_len):((i + 1) * block_enc_len)])
        dec += unpad(block_dec, block_dec_len)

    # Decode last block and unpad
    if last_block_enc_len > 0:
        block_dec = decode(
            data[(tot_block_cnt * block_enc_len):((tot_block_cnt * block_enc_len) + last_block_enc_len)])
        dec += unpad(block_dec, last_block_dec_len)

    return dec



#!/usr/bin/env python3

# Copyright © 2020-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import (
    Any, Union
)

from ..eccs import (
    IPublicKey, SLIP10Secp256k1PublicKey, validate_and_get_public_key
)
from ..cryptocurrencies import Ethereum
from ..crypto import kekkak256
from ..utils import bytes_to_string
from .iaddress import IAddress


class EthereumAddress(IAddress):

    address_prefix: str = Ethereum.PARAMS.ADDRESS_PREFIX

    @staticmethod
    def name() -> str:
        """
        Returns the name of the cryptocurrency.

        :return: The name of the cryptocurrency.
        :rtype: str
        """

        return "Ethereum"

    @staticmethod
    def checksum_encode(address: str) -> str:
        output: str = ""
        address_hash: str = bytes_to_string(
            kekkak256(address.lower())
        )
        for i, c in enumerate(address):
            if int(address_hash[i], 16) >= 8:
                output += c.upper()
            else:
                output += c
        return output

    @classmethod
    def encode(cls, public_key: Union[bytes, str, IPublicKey], **kwargs: Any) -> str:
        """
        Encodes the given public key into a Cosmos address.

        :param public_key: The public key to be encoded.
        :type public_key: Union[bytes, str, IPublicKey]
        :param kwargs: Additional keyword arguments, including "skip_checksum_encode".
        :type kwargs: Any

        :return: The encoded Cosmos address.
        :rtype: str
        """

        public_key: IPublicKey = validate_and_get_public_key(
            public_key=public_key, public_key_cls=SLIP10Secp256k1PublicKey
        )
        address: str = bytes_to_string(
            kekkak256(public_key.raw_uncompressed()[1:])
        )[24:]
        return cls.address_prefix + (
            address if kwargs.get("skip_checksum_encode", False) else cls.checksum_encode(address)
        )

    @classmethod
    def decode(cls, address: str, **kwargs: Any) -> str:
        """
        Decodes the given Cosmos address into its corresponding public key hash.

        :param address: The Cosmos address to be decoded.
        :type address: str
        :param kwargs: Additional keyword arguments, including "skip_checksum_encode".
        :type kwargs: Any

        :return: The decoded public key hash.
        :rtype: str
        """

        address_prefix_got: str = address[:len(cls.address_prefix)]
        if cls.address_prefix != address_prefix_got:
            raise ValueError(f"Invalid address_prefix (expected: {cls.address_prefix}, got: {address_prefix_got})")
        address_no_prefix = address[len(cls.address_prefix):]

        if len(address_no_prefix) != 40:
            raise ValueError(f"Invalid length (expected: {40}, got: {len(address_no_prefix)})")
        # Check checksum encoding
        if not kwargs.get("skip_checksum_encode", False) and address_no_prefix != cls.checksum_encode(address_no_prefix):
            print(address_no_prefix, cls.checksum_encode(address_no_prefix))
            raise ValueError("Invalid checksum encode")

        return address_no_prefix.lower()



#!/usr/bin/env python3

# Copyright © 2020-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from ecdsa.ecdsa import generator_256

from ...iecc import IEllipticCurveCryptography
from .point import SLIP10Nist256p1Point
from .public_key import SLIP10Nist256p1PublicKey
from .private_key import SLIP10Nist256p1PrivateKey


class SLIP10Nist256p1ECC(IEllipticCurveCryptography):

    NAME = "SLIP10-Nist256p1"
    ORDER = generator_256.order()
    GENERATOR = SLIP10Nist256p1Point(generator_256)
    POINT = SLIP10Nist256p1Point
    PUBLIC_KEY = SLIP10Nist256p1PublicKey
    PRIVATE_KEY = SLIP10Nist256p1PrivateKey
